#!/bin/bash
# Copyright 2018 Cohesity Inc.
#
# Author: Sandeep Tandekar
#
# This is a wrapper script to clone the database.
# It also appends the output at /tmp/rman_clone_db_{TASK_ID}.INFO file.
#
# Exit from the script if any of the step fails.
# clone_db.sh -R $ORACLE_HOME/bin/rman -S $ORACLE_HOME/bin/sqlplus -A cs100 -B
# cs200 -T 24:10:2018-22:42:53 -O $COHESITY_VIEW -I cohesity_4380 -V 12.1.0.2
# -H $ORACLE_HOME -J $ORACLE_BASE -F /tmp -C $COHESITY_VIEW/4376_02tgghvf_1_1
# -l $COHESITY_VIEW/4376_09tggi3d_1_1,$COHESITY_VIEW/4376_03tgghvj_1_1,$COHESITY_VIEW/4376_01tgghve_1_1,$COHESITY_VIEW/4376_02tgghvf_1_1,$COHESITY_VIEW/4376_07tggi38_1_1
# -E /tmp/switch_clause_cohesity_26901.sql -N 1024

set -e

LOG_FILE=""

# This flag is used to handle the special cases where the pattern base catalog
# doesn't work because of Oracle Bug(Oracle Doc ID 1394933.1).
# If this is set to 0(default behaviour), we will use pattern base catalog.
# If this is set to  1, we will generate a file with catalog command for
# individual backup pieces.
USE_NO_PATTERN_CATALOG=0

usage()
{
  echo "Invoke the script with following arguments :-"
  echo -e "\t -R|--rman-path                    Path to RMAN binary."
  echo -e "\t -S|--sqlplus-path                 Path to source SQLPLUS binary."
  echo -e "\t -A|--src-oracle-db-name           Oracle source database name."
  echo -e "\t -B|--dest-db-name                 Oracle destination database/SID value."
  echo -e "\t -T|--restore-time                 Restore time of database."
  echo -e "\t -O|--cohesity-view-location       Location of Cohesity view where all the files resides."
  echo -e "\t -I|--id                           Operation ID."
  echo -e "\t -F|--log-file-path                Rman log file path."
  echo -e "\t -H|--dst-oracle_home              Destination oracle home."
  echo -e "\t -J|--dst-oracle_base              Destination oracle base."
  echo -e "\t -C|--src-controlfile-path         Source control file path to be restore."
  echo -e "\t -l|--archivelog-files-prefix      Prefix for all the backuppieces."
  echo -e "\t -E|--switch-sql-file              Path to the switch SQL file."
  echo -e "\t -N|--num-data-files               Number of data files to start with."
  echo -e "\t -V|--src-database-version         Oracle source database version."
  echo -e "\t -s|--shared_pool_size             Shared Pool Size."
  echo -e "\t -t|--sga_target                   SGA Target."
  echo -e "\t -i|--fra_size                     FRA size in MBs."
  echo -e "\t -b|--bct-status                   Status of BCT(i.e. ENABLED, DISABLED, TRANSITION)."
  echo -e "\t -o|--no_open                      Restore in NOOPEN mode."
  echo -e "\t -c|--no-pattern-catalog           If set, don't use pattern based cataloging."
  echo -e "\t -d|--sleep_sec                    Amount of time in seconds to sleep after changing permission needed for AIX."
  echo -e "\t -X|--cdb_restore                  CDB Restore workflow."
  echo -e "\t -e|--custom_pfile_path            If set to valid file path, contents of this file will be appended to pfile created by agent"
  echo -e "\t -P|--script-location              SQL script location path."
  echo -e "\t -Z|--agent_pfile_path             Location of custom pfile created by agent."
  echo -e "\t -z|--pdb-list                     Comma separated list of PDB's if CDB-PDB database."
  echo -e "\t -j|--pfile_parameters_create_dir  CSV for pfile parameters for which we need to create directory and set permissions"
  echo -e "\t -k|--restore_scn                  SCN to which database needs to be restored."

}

# Shutdown the database. This routine is called as part of cleanup if clone
# fails at any stage.
shutdown_db()
{
  log_message "### Cleanup called at $1"
  log_message "### Shutting down the database as part of cleanup."
  $SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
  WHENEVER OSERROR EXIT 11
  WHENEVER SQLERROR EXIT 22
  shutdown abort;
  exit;
EOF
  exit 1
}

dump_progress() {
  # Write progress number to tmp file first, then rename temp file
  # to actual progress file for Agent code to read the file. The 'mv'
  # command is an atomic file operation on Linux. This ensures a
  # concurrent reader of the file will only read either the old version
  # or the new version of the progress file, but never partial data in
  # a new progress file.
  tmpfile="${PROGRESS_FILE}.`date +%s`.$$"
  echo "$1" > ${tmpfile}
  mv "${tmpfile}" "$PROGRESS_FILE"
}

AGENT_PFILE_PATH=""

##### Main

# Remaining alphabets for future use: D G K L M Q U W Y a f g h m n p q r u v w x y
# Please remove alphabets from the above list if you are using it.
while getopts "R:S:A:B:T:O:I:F:H:J:C:l:E:N:V:Z:s:t:i:j:b:d:ocXe:p:P:k:z:" option; do
  case $option in
    R)
      RMAN_PATH=$OPTARG;;
    S)
      SQLPLUS_PATH=$OPTARG;;
    A)
      SRC_DB_NAME=$OPTARG;;
    B)
      DEST_NEW_NAME=$OPTARG;;
    T)
      RESTORE_TIME=$OPTARG;;
    O)
      COHESITY_VIEW=$OPTARG;;
    I)
      ID=$OPTARG;;
    F)
      LOG_FILE_PATH=$OPTARG;;
    H)
      DEST_ORACLE_HOME=$OPTARG;;
    J)
      DEST_ORACLE_BASE=$OPTARG;;
    C)
      SRC_CONTROL_FILE=$OPTARG;;
    l)
      ARCHIVELOG_PREFIX=$OPTARG;;
    E)
      SWITCH_SQL_FILENAME=$OPTARG;;
    N)
      NUM_DATA_FILES=$OPTARG;;
    V)
      SRC_DB_VERSION=$OPTARG;;
    Z)
      AGENT_PFILE_PATH=$OPTARG;;
    s)
      SHARED_POOL_SIZE=$OPTARG;;
    t)
      SGA_TARGET_SIZE=$OPTARG;;
    i)
      FRA_SIZE=$OPTARG;;
    j)
      PFILE_PARAMETERS_CREATE_DIR=$OPTARG;;
    b)
      BCT_STATUS=$OPTARG;;
    o)
      NO_OPEN=1;;
    d)
      SLEEP_SEC=$OPTARG;;
    c)
      USE_NO_PATTERN_CATALOG=1;;
    X)
      CDB_RESTORE=1;;
    e)
      CUSTOM_PFILE_PATH=$OPTARG;;
    P)
      SCRIPT_LOCATION=$OPTARG;;
    z)
      PDB_LIST=$OPTARG;;
    k)
      RESTORE_SCN=$OPTARG;;
    *) echo "Invalid Option $option passed"; usage; exit 1 ;;
  esac
done

export ORACLE_HOME=$DEST_ORACLE_HOME
export ORACLE_BASE=${DEST_ORACLE_BASE}

# This variable is used during printing the log message. The script name is
# printing along with the timestamp for every log line.
SCRIPT_NAME=`basename "$0"`
source $SCRIPT_LOCATION/common_utils.sh

trap 'shutdown_db $LINENO' INT TERM

if [[ "x$LOG_FILE_PATH" == "x" ]]; then
  LOG_FILE_PATH="/tmp"
fi
LOG_FILE=${LOG_FILE_PATH}/rman_clone_db_${ID}.INFO
# Update rman log file permissions.
touch $LOG_FILE
sudo chmod 776 $LOG_FILE

# Modify how to invoke Oracle binaries if the agent is installed with 'root'
setup_invoke_oracle_binaries

# Create progress file
PROGRESS_FILE=${LOG_FILE_PATH}/${ID}_progress
touch $PROGRESS_FILE

log_message "Shell Environment:"
log_message "`env`"

# This script only uses OS authentication, just pass "/"
f_set_privilege "$DEST_ORACLE_HOME" "/"

if [ -z "$RMAN_PATH" ] || [ -z "$SQLPLUS_PATH" ] || [ -z "$SRC_DB_NAME" ] || [ -z "$SRC_DB_VERSION" ] || [ -z "$DEST_NEW_NAME" ]; then
  log_message "RMAN configuration parameters missing."
  log_message "#####$RMAN_PATH, $SQLPLUS_PATH, $SRC_DB_NAME, $SRC_DB_VERSION, $DEST_NEW_NAME";
  usage;
  exit 1;
fi

if [ -z "$RESTORE_TIME" ] || [ -z "$COHESITY_VIEW" ]; then
  log_message "RMAN script parameters missing.";
  log_message "######$COHESITY_VIEW";
  usage;
  exit 1;
fi

if [ -z "$DEST_ORACLE_HOME" ] || [ -z "$DEST_ORACLE_BASE" ]; then
  log_message "RMAN DST Oracle home/base parameter is missing.";
  log_message "#####$DEST_ORACLE_HOME, $DEST_ORACLE_BASE";
  usage;
  exit 1;
fi

if [ -z "$SWITCH_SQL_FILENAME" ] || [ -z "$SRC_CONTROL_FILE" ] || [ -z "$ARCHIVELOG_PREFIX" ]  || [ -z "$BCT_STATUS" ]; then
  log_message "datafile/redolog/archivelog_prefix/controlfile/bct_status is missing.";
  log_message "##### $SWITCH_SQL_FILENAME , $SRC_CONTROL_FILE, $ARCHIVELOG_PREFIX, $BCT_STATUS";
  usage;
  exit 1;
fi

if [ ! -f ${SWITCH_SQL_FILENAME} ]; then
  log_message "Failed to locate switch SQL file $SWITCH_SQL_FILENAME"
  exit 1;
fi

start_time="`date "+%D %X"`"
log_message "### ============== Clone started for Database ${DEST_NEW_NAME} at : ${start_time} ============="

# Change permission for datafile switch SQL file to 755.
sudo chmod 755 ${SWITCH_SQL_FILENAME}
log_message "Changed permission for ${SWITCH_SQL_FILENAME} to 755"

if [ ! -f ${DEST_ORACLE_HOME}/bin/oracle ]; then
  log_message "Unable to find ${DEST_ORACLE_HOME}/bin/oracle"
  exit 1;
fi

oracle_bin_permission=`ls -l ${DEST_ORACLE_HOME}/bin/oracle`
if [ $? -ne 0 ]; then
  log_message "Unable to find permissions for: ${DEST_ORACLE_HOME}/bin/oracle"
  exit 1;
fi

ORACLE_USER=`echo ${oracle_bin_permission} | awk '{print $3}'`
ORACLE_GROUP=`echo ${oracle_bin_permission} | awk '{print $4}'`

if [ -z ${ORACLE_USER} ] || [ -z ${ORACLE_GROUP} ]; then
  log_message "ORACLE_USER(${ORACLE_USER}) or ORACLE_GROUP(${ORACLE_GROUP}) not set."
  exit 1;
fi

# Giving permission to cohesity view.
log_message "### Running sudo chmod 1777 ${COHESITY_VIEW}"
sudo chmod 1777 ${COHESITY_VIEW};
sudo chown ${ORACLE_USER}:${ORACLE_GROUP} ${COHESITY_VIEW}
log_message "### Permission and ownership changed for ${COHESITY_VIEW}"

# Exporting PATH and LD_LIBRARY_PATH.
export LD_LIBRARY_PATH="$LD_LIBRARY_PATH:${DEST_ORACLE_HOME}/lib"
export PATH="$PATH:${DEST_ORACLE_HOME}/bin"

# Creating FRA directory
log_message "### Creating the directories as ${ORACLE_USER}:${ORACLE_GROUP}"

# Set control file destination in Oracle home directory.
CONTROL_FILES_PATH="${DEST_ORACLE_HOME}/${DEST_NEW_NAME}"
sudo mkdir -p ${CONTROL_FILES_PATH}
CONTROL_FILES="'$CONTROL_FILES_PATH/control01.ctl'"
mod_owner_group_and_permission ${CONTROL_FILES_PATH}

# Creating directory for temp files.
TMP_LOCATION="$COHESITY_VIEW/tmp/"
sudo mkdir -p ${TMP_LOCATION}
mod_owner_group_and_permission ${TMP_LOCATION}

log_message "Switch SQL file located at ${SWITCH_SQL_FILENAME}"
log_message "### Printing content of ${SWITCH_SQL_FILENAME}"
cat ${SWITCH_SQL_FILENAME} >> ${LOG_FILE}

if [ $USE_NO_PATTERN_CATALOG -eq 0 ]; then
  ARCHIVE_LOG_CLAUSE="CATALOG START WITH '${COHESITY_VIEW}/${ARCHIVELOG_PREFIX}' NOPROMPT;"
else
  filelist=""
  for file in ${COHESITY_VIEW}/${ARCHIVELOG_PREFIX}*; do
    if [ "x$filelist" = "x" ]; then
        filelist="'$file'"
    else
        filelist="$filelist,'$file'"
    fi
  done
  ARCHIVE_LOG_CLAUSE="catalog backuppiece $filelist;"
fi
log_message "### ARCHIVE_LOG_CLAUSE::: ${ARCHIVE_LOG_CLAUSE}";

# Init Automatic Diagnostic Repository (ADR).
ADR_LOCATION=${DEST_ORACLE_BASE}/admin/${DEST_NEW_NAME}
log_message "Setting up ADR at ${ADR_LOCATION}."
sudo mkdir -p ${ADR_LOCATION}
sudo mkdir -p ${ADR_LOCATION}/alert
sudo mkdir -p ${ADR_LOCATION}/trace
sudo mkdir -p ${ADR_LOCATION}/cdump
sudo mkdir -p ${ADR_LOCATION}/incident

mod_owner_group_and_permission ${ADR_LOCATION}
log_message "ADR setup done."
dump_progress "10"

echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# initora file will be created by agent.
if [ -z ${AGENT_PFILE_PATH} ]; then
  log_message "Pfile location not set."
  exit 1
else
  TMP_PFILE=$AGENT_PFILE_PATH
fi

sudo chmod 777 ${TMP_PFILE};

# If the CUSTOM_PFILE_PATH is valid, append the contents to pfile.
# Note that it should come at the end after handling all the other pfile
# options.
if [ ! -z ${CUSTOM_PFILE_PATH} ] ; then
  if [ -f ${CUSTOM_PFILE_PATH} ]; then
    log_message "Appending contents of ${CUSTOM_PFILE_PATH} to ${TMP_PFILE}"
    echo "# Appending contents from ${CUSTOM_PFILE_PATH}" >> ${TMP_PFILE}
    cat ${CUSTOM_PFILE_PATH} >> ${TMP_PFILE}
  else
    log_message "Custom pfile ${CUSTOM_PFILE_PATH} doesn't exist."
  fi
fi

sudo chmod 777 ${TMP_PFILE}
export NLS_DATE_FORMAT='DD:MM:YYYY-HH24:MI:SS'

if [[ ! -z "${PFILE_PARAMETERS_CREATE_DIR}" ]]; then
  # Create directories for some pfile parameters and also set correct permission
  create_dir_for_pfile_parameters $PFILE_PARAMETERS_CREATE_DIR $TMP_PFILE
fi

ORACLE_READ_ONLY_HOME=0
detect_read_only_oracle_home

if [ $ORACLE_READ_ONLY_HOME -eq 1 ]; then
  sudo cp ${TMP_PFILE} ${ORACLE_BASE}/dbs
  P_FILE="${ORACLE_BASE}/dbs/init${DEST_NEW_NAME}.ora";
else
  sudo cp ${TMP_PFILE} ${ORACLE_HOME}/dbs/
  P_FILE="${ORACLE_HOME}/dbs/init${DEST_NEW_NAME}.ora";
fi

mod_owner_group_and_permission ${P_FILE}
log_message "### Pfile created at ${P_FILE}, Starting database in 'nomount' mode...";
dump_progress "20"

WHICH_OS=$(uname -a | cut -d " " -f 1)

if [ ${WHICH_OS} = "AIX" ]; then
  log_message "Sleeping for ${SLEEP_SEC} seconds after attribute modification ....."
  sleep ${SLEEP_SEC}
fi

echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# After this point we want to handle the error.
set +e

# Starting new instance in no mount mode.
export ORACLE_SID=${DEST_NEW_NAME}
$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
shutdown abort;
startup  pfile='${P_FILE}' nomount;
quit;
EOF
exit_handler_with_shutdown $? "##### Startup db in 'nomount' failed."
log_message "### Instance ${DEST_NEW_NAME} started in nomount mode. Restoring control file now...";

echo "--------------------------------------------------------------------" >> ${LOG_FILE}

#Running RMAN to restore control file
$RMAN_PATH msglog=${LOG_FILE} append <<EOF
connect target '/${rman_privilege}';
set echo on
restore primary controlfile from '${SRC_CONTROL_FILE}';
alter database mount;
exit;
EOF
exit_handler_with_shutdown $? "##### Restore control file failed."
log_message "### Successfully restored control file.";

dump_progress "30"
echo "--------------------------------------------------------------------" >> ${LOG_FILE}
log_message "### Start clearing Channel info from the source backup."

# Get the channel information from SRC DB.
RMAN_CHANNEL_OUTPUT=$($RMAN_PATH <<EOF
connect target '/${rman_privilege}';
show channel;
quit;
EOF
)
exit_handler $? "##### Show channel command failed."

SRC_CHANNEL_CONFIG_STR=$(echo "$RMAN_CHANNEL_OUTPUT" | grep "CONFIGURE CHANNEL")
CLEAR_CHANNEL_LIST=""

if [[ "x${SRC_CHANNEL_CONFIG_STR}" != "x" ]];then
  CHANNEL_LIST=$( echo "$SRC_CHANNEL_CONFIG_STR" | cut -d " " -f 3)
  for i in $(eval echo ${CHANNEL_LIST});
  do
    if [ "${i}" == "DEVICE" ]
    then
      CLEAR_CHANNEL_LIST="${CLEAR_CHANNEL_LIST} CONFIGURE CHANNEL DEVICE TYPE DISK CLEAR;";
    else
      CLEAR_CHANNEL_LIST="${CLEAR_CHANNEL_LIST} CONFIGURE CHANNEL ${i} DEVICE TYPE DISK CLEAR;";
    fi
  done
else
  log_message "No Channel configuration found in the src DB"
fi

# Clear the channel information from the source DB which came with the backup
# control file.
if [[ "x${CLEAR_CHANNEL_LIST}" != "x" ]];then
  log_message "Running following command to clear channel config from backup."
  log_message "${CLEAR_CHANNEL_LIST}"
  $RMAN_PATH msglog=${LOG_FILE} append <<EOF
  connect target '/${rman_privilege}';
  set echo on;
  ${CLEAR_CHANNEL_LIST}
  quit;
EOF
  exit_handler_with_shutdown $? "##### Clear channel command failed."
  log_message "### Successfully Cleared channel configuration from backup."
fi

log_message "### Clear Channel block completed."
dump_progress "40"

log_message "will force to turn flashback off before recovery"

UNTIL_CLAUSE="SET UNTIL TIME \"TO_DATE('$RESTORE_TIME','DD:MM:YYYY-HH24:MI:SS')\";"

if [ ! -z $RESTORE_SCN ]; then
  log_message "Restoring database to SCN: $RESTORE_SCN instead of time: $RESTORE_TIME";
  UNTIL_CLAUSE="SET UNTIL SCN $RESTORE_SCN;"
fi

# Running RMAN to catalog and recover.
# For the current logic, we will catalog all the datafiles and execute switch
# database to copy, This will switch all the datafiles to copy.
$RMAN_PATH msglog=${LOG_FILE} append <<EOF
connect target '/${rman_privilege}';
set echo on;
SET COMMAND ID TO '$ID';
@${SWITCH_SQL_FILENAME}
${ARCHIVE_LOG_CLAUSE}
crosscheck archivelog all;
delete NOPROMPT expired archivelog all;
crosscheck datafilecopy all;
delete NOPROMPT expired datafilecopy all;
crosscheck backup;
switch database to copy;
EOF
exit_handler_with_shutdown $? "##### Switch database to copy failed for $ORACLE_SID at: ${RESTORE_TIME}"

# Once the datafiles are switched successfully, we will be
# able to recover the copy/datafiles on the clone view.
$RMAN_PATH msglog=${LOG_FILE} append <<EOF
connect target '/${rman_privilege}';
set echo on;
SET COMMAND ID TO '$ID';
delete NOPROMPT expired backup;
sql 'alter database flashback off';
list backup;
list datafilecopy all;
RUN {
  ${UNTIL_CLAUSE}
  RECOVER database;
}
exit;
EOF
exit_handler_with_shutdown $? "##### Clone copy recovery failed on $ORACLE_SID at: ${RESTORE_TIME}"
log_message "### Clone copy recovery successful. Creating sql scripts now...";
dump_progress "50"

echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# Generate controlfile and tempfile creation sql scripts.
export ORACLE_SID=${DEST_NEW_NAME}

# Generate the sql script for control file creation.
generate_create_control_file_sql_script

# Generate SQL script to add temp files into the database.
generate_temp_file_sql_script

log_message "Successfully generated the sql scripts, creating sqfile now."

echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# Creating spfile.
$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
create spfile from pfile;
shutdown abort;
exit;
EOF
exit_handler_with_shutdown $? "##### SPFile creation failed."
log_message "### SPFile created successfully.";

$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
startup nomount;
ALTER SYSTEM SET DB_NAME=${DEST_NEW_NAME} SCOPE=SPFILE;
shutdown abort;
exit;
EOF
exit_handler_with_shutdown $? "##### Reset DB_NAME failed."
log_message "### Reset DB_NAME successfully.";

$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
startup nomount;
create pfile from spfile;
shutdown abort;
exit;
EOF
exit_handler_with_shutdown $? "##### Refresh SPFILE failed."
log_message "### Refresh SPFILE successfully.";

# Remove old control file.
control_files_values=$(grep -i "CONTROL_FILES=" $TMP_PFILE | tail -n 1 | cut -d'=' -f2)
# remove ' and " if present
control_files_values=$(echo "$control_files_values" | tr -d "'" | tr -d '"')
IFS=',' read -r -a control_files <<< "$control_files_values"
echo ${control_files[*]}
for control_file in "${control_files[@]}"
do
  if [[ $control_file == +*  ]]; then
    log_message "Not removing control file in asm: $control_file"
  else
    log_message "Removing old control file: $control_file"
    WHITELISTED_PATH_LIST+=( "$control_file" )
    safe_rm "__USE_SUDO__" "$control_file"
  fi
done

dump_progress "70"
echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# Check to leave the DB is open mode. if no open (Leave db is recovery mode) is
# selected, then the database will be left in mount mode.
if [ -z ${NO_OPEN} ]; then
  OPEN_CLAUSE="ALTER DATABASE OPEN RESETLOGS;"
else
  OPEN_CLAUSE=""
fi
log_message "###OPEN_CLAUSE:: ${OPEN_CLAUSE}"

# Set the undo management to auto for the clone, this is required to open the
# pdbs, and if not specifed, the pluggable db's cannot be opened.
if [ ! -z ${CDB_RESTORE} ]; then
  CDB_CLAUSE="ALTER SYSTEM SET UNDO_MANAGEMENT=AUTO SCOPE=SPFILE;"
  if [ -z ${NO_OPEN} ]; then
    CDB_OPEN_CLAUSE="ALTER PLUGGABLE DATABASE ALL OPEN;"
  fi
fi

#Enable BCT if it was configured on the source.
if [[ "$BCT_STATUS" == "ENABLED" ]]; then
  BCT_CLAUSE="ALTER DATABASE ENABLE BLOCK CHANGE TRACKING;"
fi

echo "--------------------------------------------------------------------" >> ${LOG_FILE}
# Create new control file, open the DB and run CDB specific workflow.
# Create number of tempfiles based on the source.
$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
STARTUP NOMOUNT;
ALTER SYSTEM SET thread=0 SCOPE=BOTH;
@${control_file_sql}
ALTER DATABASE OPEN RESETLOGS;
exit;
EOF

# Remember the status of cloning
clone_result=$?

# Best try effort - add tempfile and setup BCT
$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
${BCT_CLAUSE}
@${tempfile_sql}
shutdown immediate;
exit;
EOF

[ $? -ne 0 ] && log_message "failed to ADD TEMPFILE or set Block Change Tracking"

exit_handler_with_shutdown $clone_result "##### Clone failed to create control file and open resetlogs (SID=$ORACLE_SID, Point-in-time=${RESTORE_TIME})"
log_message "Control file created, followed by successful open resetlogs."

$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
startup mount;
exit;
EOF

exit_handler_with_shutdown $? "##### Clone failed to startup after successful open resetlogs (SID=$ORACLE_SID, Point-in-time=${RESTORE_TIME})"

echo "--------------------------------------------------------------------" >> ${LOG_FILE}
# Change the DBID for the clone.
NID_LOG_FILE=${LOG_FILE_PATH}/db_rename_${ID}.INFO

log_message "### Renaming the database using 'nid' utility for user : ${ORACLE_USER}"
${NID_PATH} TARGET=/ LOGFILE=${NID_LOG_FILE};
exit_handler_with_shutdown $? "##### Rename failed."
log_message "### Rename using 'nid' successful. Opening the database with RESETLOGS...";

$SQLPLUS_PATH / $sqlplus_privilege << EOF >> ${LOG_FILE}
WHENEVER OSERROR EXIT 11;
WHENEVER SQLERROR EXIT 22;
STARTUP MOUNT;
${OPEN_CLAUSE}
${CDB_OPEN_CLAUSE}
exit;
EOF

exit_handler_with_shutdown $? "##### Clone failed (SID=$ORACLE_SID, Point-in-time=${RESTORE_TIME})"
log_message "### Database ${DEST_NEW_NAME} started.";
log_message "##### Clone completed (SID=$ORACLE_SID, Point-in-time=${RESTORE_TIME})";
dump_progress "90"
# Updating oratab file.
echo "$DEST_NEW_NAME:$ORACLE_HOME:N" | sudo tee -a /etc/oratab;

exit 0
