#!/bin/bash
# Copyright 2018 Cohesity Inc.
#
# Author: Mahesh.K
#
# Common util shell script to hold all the sql/oracle common functions which can
# be used in multiple places.

# RMAN backup type.
export IMAGECOPY_BACKUP="1";
export BACKUPSETS_BACKUP="2";
export SBT_BACKUP="3";

log_message()
{
  msg=$*
  TS=`date "+I%m%d %H:%M:%S"`
  if [ "$LOG_FILE" == "" ]; then
    echo "$TS [$SCRIPT_NAME] ${msg}"
  else
    echo "$TS [$SCRIPT_NAME] ${msg}" >> ${LOG_FILE}
  fi
}

# This function logs the error and exit with error code if caller routine failed.
exit_handler() {
  status=$1;
  if [ $status -eq 0 ]; then
    return;
  fi

  # Doing cleanup.
  log_message "RMAN-9999 $2";
  exit $status;
}

# This function logs the error and calls the shutdown routine before exiting
# with error code if caller routine failed.
exit_handler_with_shutdown() {
  status=$1;
  if [ $status -eq 0 ]; then
    return;
  fi

  # The caller needs to implement the shutdown_db routine.
  shutdown_db;

  # Doing cleanup.
  log_message "RMAN-9999 $2";
  exit $status;
}

# This function use an Oracle version based logic to determine if we should use
# SYSDBA or SYSBACKUP to connect to the Oracle database. A connection test is
# done and adjustment made to fallback to use the default SYSDBA way if
# SYSBACKUP is not setup to be used in the environment.
f_set_privilege() {
  local oracle_home="$1"
  local local_connect_string="$2"

  if [ ! -f $oracle_home/bin/sqlplus ]; then
      log_message "Unable to find $oracle_home/bin/sqlplus"
      exit 1;
  fi

  local rman_path=$RMAN_PATH
  if [ "X${RMAN_PATH}" == "X" ]; then
      if [ ! -f $oracle_home/bin/rman ]; then
        log_message "Unable to find $oracle_home/bin/rman"
        exit 1;
      fi
      # Fallback to use default rman binary location per Oracle Home
      rman_path="$oracle_home/bin/rman"
  fi

  local use_sysbackup=0
  local oracle_sqlplus_version=$($oracle_home/bin/sqlplus -V | grep Release | cut -d " " -f 3)
  IFS='.' read -ra ADDR <<< "$oracle_sqlplus_version"
  if [ ${ADDR[0]} -gt 12 ] || { [ ${ADDR[0]} -eq 12 ] && [ ${ADDR[1]} -ge 1 ]; }; then
    # Test if we can connect with SYSBACKUP, if not, fallback automatically to
    # sysdba. Some customer environment that is >=12c may not be able to
    # set up SYSBACKUP in their OS authentication or DB authentication. For
    # example, if SYSBACKUP OS group is not specified when the oracle image is
    # built previously, this will require production database down time to
    # add the SYSBACKUP OS group and relink of the Oracle binary. Customer
    # may not be able to do that. Cohesity should not impose SYSBACKUP as a
    # hard requirement with database >=12c.
    #
    # The test is being done using RMAN to connect. The result should also be
    # applicable to the SQLPLUS case.
    set +e
    $rman_path <<EOF > /dev/null
      connect target '$local_connect_string as sysbackup';
EOF
    [ $? = 0 ] && use_sysbackup=1
    set -e
  fi

  # For PDB level restore, we can't use sysbackup privileges as the restore
  # tries to drop few tablespaces which are not permitted n sysbackup,
  # subsequently causing the restore to fail.
  if [ ${use_sysbackup} -eq 1 ] && [ -z ${PDB_LIST} ]; then
    sqlplus_privilege="as sysbackup"
    rman_privilege=" as sysbackup"
  else
    sqlplus_privilege="as sysdba"
    # 11c Not support 'as sysdba' when using RMAN
    rman_privilege=""
  fi

  # Check if LOG_FILE exists as some operation rely on stdout
  if [ "$LOG_FILE" != "" ]; then
    log_message "### RMAN or SQLPLUS connect privilege to use: $sqlplus_privilege."
  fi
}

# Get the max log per undo group.
get_max_log_member() {
  vlogmembers="x\$kccdi"
  local VarName="$1"
  local local_connect_string="$2"
  retVal=`$SQLPLUS_PATH -s /nolog << EOF
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect ${local_connect_string} $sqlplus_privilege
  set head off
  set feedback off
  set pagesize 5000
  set linesize 30000
  Select dimlm from $vlogmembers;
  EXIT;
EOF
`
  exit_handler  $? "##### Failed to fetch the max log member value."
  eval $VarName=\"${retVal}\"
}

# Get the value for a parameter and store it in a variable.
get_db_property_value() {
  local VarName="$2"
  local local_connect_string="$3"
  retVal=`$SQLPLUS_PATH -s /nolog << EOF
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect ${local_connect_string} $sqlplus_privilege
  set head off
  set feedback off
  set pagesize 5000
  set linesize 30000
  select PROPERTY_VALUE
    from database_properties
    where PROPERTY_NAME= '$1';
  EXIT;
EOF
`
  exit_handler  $? "##### Failed to fetch the $1 value from oracle database."
  eval $VarName=\"${retVal}\"
}

# Get the record value from v$controlfile_record_section view.
get_vctl_value() {
  vctl="V\$controlfile_record_section"
  local VarName="$2"
  local local_connect_string="$3"
  retVal=`$SQLPLUS_PATH -s /nolog << EOF
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect ${local_connect_string} $sqlplus_privilege
  set head off
  set feedback off
  set pagesize 5000
  set linesize 30000
  select records_total
    from  $vctl
    where TYPE= '$1';
  EXIT;
EOF
`
  exit_handler  $? "##### Failed to fetch the $1 value from oracle database."
  eval $VarName=\"${retVal}\"
}

# Create control file function.

### Recreate controlfile
# Okay to use "/" as connect string here, as this is used only in recovery.
generate_create_control_file_sql_script() {
  UNDO_LOG_PATH="'${COHESITY_VIEW}/'"
  control_file_sql=/tmp/control_file_"${ID}".sql
  WHITELISTED_PATH_LIST+=( "${control_file_sql}" )
  temp_control_file_sql=/tmp/temp_control_file_"${ID}".sql
  WHITELISTED_PATH_LIST+=( "${temp_control_file_sql}" )
  safe_rm_force_recursive "$control_file_sql"
  safe_rm_force_recursive "$temp_control_file_sql"

  log_message "Generating the control file creation script."

  ####  Getting database properties for controlfile Generation
  # Here, we will query the datafile section and undo sections from the control
  # file and then create a sql script to create the control file with required
  # parameters.
  # Redo log files will be set under $UNDO_LOG_PATH with modified filenames
  # If original redo log file has extension, then "_${group id}" will be appended
  # before extension part, othwerwise "_${group id}.log" will be appended
  # For example, here are original redo log files
  # /u01/oracle/redo01.log
  # /u01/oracle/redo02.log
  # If destination redo log path is /u02/oracle, after conversion, files are:
  # /u02/oracle/redo01_1.log
  # /u02/oracle/redo02_2.log
  # If original redo log files do not have extension, like
  # /u01/oracle/redo1
  # /u01/oracle/redo2
  # After conversion, files are:
  # /u01/oracle/redo1_1.log
  # /u01/oracle/redo2_2.log
  $SQLPLUS_PATH -s /nolog << EOF > $temp_control_file_sql
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect / $sqlplus_privilege
  SET VERIFY OFF;
  SET ECHO OFF;
  SET SERVEROUTPUT ON;
  SET HEADING OFF;
  SET FEEDBACK OFF;
  SET LINE 32000;
  select 'CREATE CONTROLFILE SET DATABASE ' || '"' ||
  db_unique_name || '"' || ' RESETLOGS  ARCHIVELOG '
  from v\$database;
  Select 'MAXDATAFILES ' || records_total
  from v\$controlfile_record_section
  where type = 'DATAFILE';
  Select 'MAXLOGHISTORY ' || records_total
  from v\$controlfile_record_section
  where type = 'LOG HISTORY';
  Select 'MAXINSTANCES ' || records_total
  from v\$controlfile_record_section
  where type = 'DATABASE';
  Select 'MAXLOGFILES ' || records_total
  from v\$controlfile_record_section
  where type = 'REDO LOG';
  Select 'MAXLOGMEMBERS ' || dimlm from x\$kccdi;
  select 'LOGFILE' from dual;
  select 'GROUP ' || a.GROUP# || '  ' ||  '(''' || LISTAGG($UNDO_LOG_PATH ||
  case when instr(substr(a.member,instr(a.member,'/',-1,1)+1), '.') = 0
  then substr(a.member,instr(a.member,'/',-1,1)+1) || '_' || rownum || '.log'
  else replace(substr(a.member,instr(a.member,'/',-1,1)+1), '.', '_' || rownum || '.') end,''',''')
  WITHIN GROUP(order by a.GROUP#) || ''')'||
  ' size ' || TRUNC(b.bytes/1024/1024) || 'M, '
  from v\$logfile a , v\$log b
  where a.group# <= (select max(group#) from v\$log ) and a.group#=b.group#
  group by a.group#  ,b.bytes;
  select 'DATAFILE' from dual;
  select '''' || name  || '''' || ','
  from v\$datafile
  where file# <(select max(file#) from v\$datafile);
  select '''' || name  || ''''
  from v\$datafile
  where file# =  ( select max(file#) from v\$datafile);
  select 'CHARACTER SET ' || value
  from v\$nls_parameters
  where PARAMETER = 'NLS_CHARACTERSET';
  select ';' from dual;
EOF
  exit_handler $? "Failed to generate control file sql script."
  log_message "Control file sql script successfully created."
  ### Format control file
  sed '/^$/d'  $temp_control_file_sql > $control_file_sql
  sudo chmod 775 ${control_file_sql}
  safe_rm_force_recursive "$temp_control_file_sql"
}

# Okay to use "/" as connect string here, as this is used only in recovery.
generate_temp_file_sql_script() {
  tempfile_sql=/tmp/temp_file_"${ID}".sql
  log_message "Generating the Temp file sql script:: $tempfile_sql"

  # If clone is being performed on CDB, We need to add an additional clause for
  # getting the temp files. as there can be multiple temp files with the same ID
  # for different PDBs.
  if [ ! -z ${CDB_RESTORE} ]; then
    CDB_TEMP_FILE_CLAUSE=" and a.con_id = b.con_id"
  fi

  # SQL Script to create temp files.
  # Query the tempfile view from the control file and then formulate an sql script
  # to create the list of tempfile.
  $SQLPLUS_PATH -s /nolog << EOF > $tempfile_sql
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect / $sqlplus_privilege
  SET VERIFY OFF;
  SET ECHO OFF;
  SET SERVEROUTPUT ON;
  SET HEADING OFF;
  SET FEEDBACK OFF;
  SET LINE 32000;
  select 'alter tablespace '|| b.name ||' add tempfile ''${TMP_LOCATION}/temp' ||
  a.File# || '.dbf'' size ' ||
  TRUNC(a.CREATE_BYTES/1024/1024) || 'M;'
  from v\$tempfile a, v\$tablespace b
  where a.ts# = b.ts#
  ${CDB_TEMP_FILE_CLAUSE};
EOF
  exit_handler $? "Failed to generate temp file sql script."
  sudo chmod 775 ${tempfile_sql}
  log_message "Temp file sql script successfully created."
}

# Get the value for a parameter and store it in a variable.
get_parameter_value() {
  local VarName="$2"
  local local_connect_string="$3"
  PARAMETER="v\$PARAMETER"
  retVal=`$SQLPLUS_PATH -s /nolog << EOF
  connect ${local_connect_string} as sysdba
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  set head off
  set feedback off
  set pagesize 5000
  set linesize 30000
  select VALUE
    from $PARAMETER
    where NAME = '$1';
  EXIT;
EOF
`
  status=$?;
  if [ $status -ne 0 ]; then
    log_message "Failed to query following parameter: $1"
    unset retVal
  fi
  eval $VarName=\"${retVal}\"
}

# Delete files from disk.
delete_files() {
  IFS=',' read -ra file_list <<< "$1"
  for files in "${file_list[@]}"
    do
      # ASM Check.
      # TODO(Mahesh): Logic to delete files from ASM.
      if [[ ${files} != +*  ]] || [ -f ${files} ]; then
        log_message "Deleting file: $files"
        safe_rm_force "__USE_SUDO__" "$files"
      fi
    done
}

# Function to cleanup the control files, spfiles and the datafiles for a restore
# failure or cancellation scenario.
cleanup_restore_alternate() {
  if [ ! -z $RMAN_COMPLETED ] && [ $RMAN_COMPLETED -ne 0 ]; then
    log_message "RMAN Completed, dropping the db."
    $SQLPLUS_PATH / as sysdba << EOF >> ${LOG_FILE}
      shutdown abort;
      startup mount exclusive restrict;
      drop database;
      quit;
EOF
  else
    get_parameter_value spfile 'SPFILE' "/"
    get_parameter_value controlfiles 'CONTROL_FILES' "/"
    $SQLPLUS_PATH / as sysdba << EOF >> ${LOG_FILE}
      shutdown abort;
      quit;
EOF
    log_message "Manually cleaning up the files. $DATAFILE_LOCATION, $SPFILE, $CONTROL_FILES"
    # Delete Data directory only if non asm.
    if [ ! -z $NON_ASM ]; then
      log_message "Deleting datafile location: $DATAFILE_LOCATION"
      safe_rm_force_recursive "__USE_SUDO__" "$DATAFILE_LOCATION/*"
    fi
    delete_files $CONTROL_FILES
    delete_files $SPFILE
  fi
}

#
# This function checks if the agent is installed with 'root' or any non-root OS
# user. It then modifies how to invoke the following Oracle binaries:
#   - SQLPLUS
#   - RMAN
#   - NID
# accordingly, to allow in both cases of 'root' or 'non-root' running Agent,
# the script will be able to use OS authentication to execute either the
# sqlplus or rman command.
#
# By default, this function looks for the owner of the Oracle executable and
# use that as the user to sudo into. Oracle binary owner can always gain access
# as SYSDBA through OS authentication.
#
# The following shell variables are referenced as input to this function:
#    - AGENT_INSTALL_USER: (optionally) passed into the script from Agent gflag
#    - ORACLE_HOME: (requiredd) the current ORACLE_HOME of interest
#
setup_invoke_oracle_binaries() {
  if [ -z ${ORACLE_HOME} ]; then
    log_message "######Error: setup_invoke_oracle_binaries() must have ORACLE_HOME defined!"
    exit 1;
  fi

  # Initialize SQLPLUS_PATH and RMAN_PATH with default values, if it
  # does not already have a value.
  [ -z "${SQLPLUS_PATH}" ] && export SQLPLUS_PATH="${ORACLE_HOME}/bin/sqlplus"
  [ -z "${RMAN_PATH}" ] && export RMAN_PATH="${ORACLE_HOME}/bin/rman"
  export NID_PATH="${ORACLE_HOME}/bin/nid"

  # Is the agent running with 'root'? If yes, change SQLPLUS_PATH,
  # RMAN_PATH and NID_PATH to modify the way how we will invoke these Oracle
  # binaries.
  #
  # DO NOT MODIFY HOW TO INVOKE ORACLE BINRARIES IF THIS IS AIX. AIX agent has
  # its own way of handling 'root' installed Agent.
  if [ $(whoami) = 'root' ] && [ $(uname -a | cut -d " " -f 1) != "AIX" ]; then
    # User passed in $AGENT_ORACLE_USER, if there is one. Otherwise, use
    # oracle binary file owner.

    [ -z ${AGENT_ORACLE_USER} ] && export AGENT_ORACLE_USER=$(ls -l $ORACLE_HOME/bin/oracle | awk '{print $3}')

    export SQLPLUS_PATH="sudo -u ${AGENT_ORACLE_USER} -E ${ORACLE_HOME}/bin/sqlplus"
    export RMAN_PATH="sudo -u ${AGENT_ORACLE_USER} -E ${ORACLE_HOME}/bin/rman"
    export NID_PATH="sudo -u ${AGENT_ORACLE_USER} -E ${ORACLE_HOME}/bin/nid"
  fi
}

# Adds entry to command path hashmap.
#
# Arguments:
#  arg1: Key
#  arg2: Value
#
f_insert_cmd_path() {
  eval cmd_path_"$1"='$2'
}

# Go through each command in RequiredCommandsArray and find out its full path.
# If we failed to find full path of any command, we exit right away.
# Full paths found are populated in global hashmap with f_hput() API which is
# explained above.
f_find_cmd_full_path_checked() {
  # This array contains all commands for which we need to find full path.
  local RequiredCommandsArray=(
    "chmod"
    "chown"
    "mkdir"
    "touch"
  )

  for key in "${RequiredCommandsArray[@]}"
  do
    local my_cmd=$key

    # Find out full path of command.
    local cmd_path=$(whereis $my_cmd | awk '{ print $2 }')
    if [[ "x$cmd_path" == "x" ]] ; then
      log_message "Error: Failed to find full path of ${my_cmd}."
      exit 1
    fi
    f_insert_cmd_path "$key" "$cmd_path"
  done
}


# Function to detect the oracle read only oracle_home configuration.
# if read only oracle home is present, oracle writes the entry in the
# $ORACLE_HOME/install/orabasetab file with the entry set to Y, we will detect
# it and take appropriate action.

detect_read_only_oracle_home() {
  if [ -f $ORACLE_HOME/install/orabasetab ]; then
    log_message "Oracle base tab exists"
    ORACLE_HOME_DIR_NAME=`dirname $ORACLE_HOME`
    ORACLE_BASE_TAB_LOC=$ORACLE_HOME/install/orabasetab
    ORACLE_READ_ONLY_HOME_STR=`grep $ORACLE_HOME_DIR_NAME $ORACLE_BASE_TAB_LOC | awk '{split($0,array,":")}  END{print array[4]}'`
    if [ "$ORACLE_READ_ONLY_HOME_STR" = "Y" ]; then
      ORACLE_READ_ONLY_HOME=1
    fi
  fi
}

# Method to check if given file lies under a given directory path.
# In case of symlink directories, it checks for the entire path.
# Return 0 if file is under given path, else 1.
f_check_file_lies_under_given_dir() {
  local file_path=$1
  local dir_path=$2
  # Check if given directory is a symlink, if so get full path.
  test -L $dir_path
  if [[ "$?" -eq "0" ]]; then
    local dir_path_provided=$dir_path
    dir_path=`readlink $dir_path`
    log_message "Updated symlink path from $dir_path_provided to $dir_path"
  fi

  local parent_dir_path=`dirname $file_path`
  test -L $parent_dir_path
  if [[ "$?" -eq "0" ]]; then
    local parent_dir_path_provided=$parent_dir_path
    parent_dir_path=`readlink $parent_dir_path`
    log_message "Updated symlink path from $parent_dir_path_provided to $parent_dir_path"
  fi

  # Check if file lies under given directory.
  if [[ "$parent_dir_path" == "$dir_path" ]]; then
    return 0
  fi

  return 1
}

# Get local time from UTC time.
get_local_time() {
  local VarName="$1"
  local local_connect_string="$3"
  retVal=`$SQLPLUS_PATH -s /nolog << EOF
  WHENEVER OSERROR EXIT 11;
  WHENEVER SQLERROR EXIT 22;
  connect ${local_connect_string} $sqlplus_privilege
  set head off
  set feedback off
  set pagesize 5000
  set linesize 30000
  alter session set NLS_DATE_FORMAT = 'DD:MM:YYYY-HH24:MI:SS';
  SELECT CAST(FROM_TZ(CAST(TO_DATE('$2','DD:MM:YYYY-HH24:MI:SS') AS TIMESTAMP), 'UTC') AT LOCAL AS DATE) from dual;
  EXIT;
EOF
`
  exit_handler $? "##### Failed to convert UTC time $2 to local time, output: ${retVal}"
  eval $VarName=\"${retVal}\"
}

# Helper method to check if given path is ASM
f_check_is_asm_path() {
  IS_ASM="false"

  if [[ $1 == +* ]] && [ ! -d $1 ]; then
    IS_ASM="true"
  fi
}

# Take an input directory, work backward until we find the first parent directory
# that does not exist currently. E.g. when we create a new Flash Recovery Area
# directory for alternate restore, the pfile parameter may speciy a directory like
# /u01/app/oracle/product/12.1.0.2/db_1/fast_recovery_area/db2, where the base dir
# /u01/app/oracle/product/12.1.0.2/db_1/ exists, by our retore workflow will need
# to create both the 'fast_recovery_area' and 'db2' subdirectory. This function
# returns the path to the first directory that we will create (i.e. that does not
# currently exist). This will then be used as the root directory to do recurive
# changing of owner and group, as well as permission.
#
# Note: input first argument cannot be a relative directory
get_base_first_new_directory() {
  local curdir=$1
  local prevdir=$1
  while [ ! -d $curdir ]; do
    prevdir=$curdir;
    curdir=`dirname $curdir 2>/dev/null`;
    [ X${curdir} == "X" ] && break;
  done;
  echo $prevdir
}

# Set correct permission and group for a directory
mod_owner_group_and_permission() {
  local dirpath=$1
  local ORACLE_USER=`echo ${oracle_bin_permission} | awk '{print $3}'`
  local ORACLE_GROUP=`echo ${oracle_bin_permission} | awk '{print $4}'`
  sudo chmod -R 775 ${dirpath}
  sudo chown -R ${ORACLE_USER}:${ORACLE_GROUP} ${dirpath}
  log_message "### Permission 775 and ownership ${ORACLE_USER}:${ORACLE_GROUP} changed for ${dirpath}"
}

# Create directories for some pfile parameters and also set correct permission
create_dir_for_pfile_parameters() {
  local PFILE_PARAMETERS_CREATE_DIR=$1
  local TMP_PFILE=$2
  if [ -z $TMP_PFILE ]; then
    log_message "Path to pfile not passed to create_dir_for_pfile_parameters function."
    exit 1
  fi
  if [[ ! -z "${PFILE_PARAMETERS_CREATE_DIR}" ]]; then
    IFS=',' read -r -a PARAMETER_ARRAY <<< "$PFILE_PARAMETERS_CREATE_DIR"
    for PARAMETER in "${PARAMETER_ARRAY[@]}"
    do
      log_message "Searching for value of ${PARAMETER} in pfile for directory creation."
      PARAMETER_VALUE=$(grep -i "${PARAMETER}=" ${TMP_PFILE} | tail -n 1 | cut -d'=' -f2)
      if [[ ! -z ${PARAMETER_VALUE} ]]; then
        # remove ' and " if present
        PARAMETER_VALUE=$(echo "$PARAMETER_VALUE" | tr -d "'" | tr -d '"')
        if [[ $PARAMETER_VALUE == +*  ]]; then
          log_message "Cannot create directory for pfile parameter ${PARAMETER} as its value $PARAMETER_VALUE is in asm."
        else
          log_message "Creating directory $PARAMETER_VALUE and setting correct permission of it for pfile parameter ${PARAMETER}."
          local dir1=`get_base_first_new_directory $PARAMETER_VALUE`
          sudo mkdir -p $PARAMETER_VALUE
          mod_owner_group_and_permission $dir1
        fi
      fi
    done
  else
    log_message "Pfile parameters not passed to create_dir_for_pfile_parameters function."
    exit 1
  fi
}

#------------------------------------------------------------------------------

# List of paths tht cannot be deleted.
BLACKLIST_PATH_LIST=("." ".." "/*" "/" "*" ".*" "/etc" "/etc/fstab" "/home" "/tmp")
# Array containing path pattern based on which files/dirs can be deleted.
# NOTE: In case the script is taking a down-path, then it is parent script's
# responisiblity to clear this to avoid any old/garbage enteries.
WHITELISTED_PATH_LIST=()

# Create directory and append entry in CREATED_PATH_MAP.
safe_mkdir() {
  local mkdir_cmd="mkdir"
  if [ "$1" == "__USE_SUDO__" ]; then
    mkdir_cmd="sudo $mkdir_cmd"
    shift
  fi

  for path in "$@"
  do
    if [ ! -d "${path}" ]; then
      WHITELISTED_PATH_LIST+=( "${path}" )
      log_message "Creating dir, running command: ${mkdir_cmd} -p ${path}"
      ${mkdir_cmd} -p ${path}
      local err_code=$?
      if [[ $err_code -ne 0 ]]; then
        log_message "Failed command: ${mkdir_cmd} -p ${path}"
        return $err_code
      fi
    fi
  done
}

# Wrapper function for "rm -r" and "rm" command.
# Deletes a given path only if its not in critical path list and
# and has parent dir in WHITELISTED_PATH_LIST
# If first argument is __FORCE__ then it will skips paths not satisfying
# deletion criteria else it will return error if any path does not satisfy
# deletion criteria.
rm_wrapper() {
  local rm_arg=""
  local continue_on_error=false
  if [ "$1" == "__FORCE__" ]; then
    rm_arg="${rm_arg}f"
    continue_on_error=true
    shift
  fi
  if [ "$1" == "__RECURSIVE__" ]; then
    rm_arg="${rm_arg}r"
    shift
  fi
  if [[ "x${rm_arg}" != "x" ]]; then
    rm_arg="-${rm_arg}"
  fi

  local sudo_cmd=""
  if [ "$1" == "__USE_SUDO__" ]; then
    sudo_cmd="sudo"
    shift
  fi

  for path in "$@"
  do
    local skip_delete=false
    local normalized_path=${path}
    if [[ ! -L "${path}" ]]; then
      # Normalize the given path to remove extra "///", "/././", etc
      # Example "/..////././." will be normalized to "/"
      # Before performing the deletion on the path, we want to normalize it so
      # that we can filter out some error cases like empty paths, root dirs,
      # home dirs etc.
      local normalized_path=`readlink -m "${path}"`
    fi
    log_message "Attempting to delete path:${path}, normalized_path: ${normalized_path}"

    if [ "x${normalized_path}" == "x" ]; then
      if [[ "${continue_on_error}" == "true" ]]; then
        log_message "Skipping empty path deletion."
        continue
      else
        log_message "Cannot delete empty path."
        return 1
      fi
    fi

    # Look for critical paths.
    for critical_path in "${BLACKLIST_PATH_LIST[@]}"
    do
      if [[ "${normalized_path}" == "${critical_path}" ]]; then
        if [[ "${continue_on_error}" == "true" ]]; then
          log_message "Skipping critical path deletion, path provided: ${path}"
          skip_delete=true
          break
        else
          log_message "Cannot delete critical path: ${path}"
          return 1
        fi
      fi
    done
    if [[ "${skip_delete}" == "true" ]]; then
      continue
    fi

    # Check if path is there in WHITELISTED_PATH_LIST.
    for whitelisted_path in "${WHITELISTED_PATH_LIST[@]}"
    do
      if [[ "${path}" =~ ${whitelisted_path} ]]; then
        log_message "Removing path, running command: ${sudo_cmd} rm ${rm_arg} ${path}"
        ${sudo_cmd} rm ${rm_arg} ${path}
        local err_code=$?
        if [[ $err_code -ne 0 ]]; then
          log_message "Failed command: ${sudo_cmd} rm ${rm_arg} ${path}"
          return $err_code
        fi
        break
      fi
    done
  done
}

# Safe function for "rm" command.
# Deletes a given path only if its not in critical path list and
# and has parent dir in WHITELISTED_PATH_LIST
# Fails if any path does not satisfy deletion criteria.
safe_rm() {
  rm_wrapper "$@"
}

# Safe function for "rm -r" command.
# Deletes a given path only if its not in critical path list and
# and has parent dir in WHITELISTED_PATH_LIST
# Fails if any path does not satisfy deletion criteria.
safe_rm_recursive() {
  rm_wrapper "__RECURSIVE__" "$@"
}

# Safe function for "rm -f" command.
# Deletes a given path only if its not in critical path list and
# and has parent dir in WHITELISTED_PATH_LIST
# Skips paths not satisfying deletion criteria.
safe_rm_force() {
  rm_wrapper "__FORCE__" "$@"
}

# Safe function for "rm -rf" command.
# Deletes a given path only if its not in critical path list and
# and has parent dir in WHITELISTED_PATH_LIST
# Skips paths not satisfying deletion criteria.
safe_rm_force_recursive() {
  rm_wrapper "__FORCE__" "__RECURSIVE__" "$@"
}
