#!/bin/bash
# Copyright 2019 Cohesity Inc.
#
# Author: Steve Lee
#
# This script performs an RMAN level 0 (full) or level 1 (incremental) backup.
# For BackupSet backup, the passed in 'BACKUP_LEVEL' determines the backup
# level to be performed. For ImageCopy backup, Magneto controls the starting
# state of the backup location. If there are no datafile image copies at the
# backup destination location, Oracle will perform a Level 0 full backup;
# otherwise, a level 1 incremental backup will be performed.

#
# Only parse options that are of interest to this backup script.
#

# Which OS Cohesity agent is running on.
OS=$(uname -a | cut -d " " -f 1)

if [ ${OS} = "AIX" ]; then
  # TODO(aabhas): find replacement for readlink on AIX.
  current_dir=$(dirname $0)
else
  current_dir=$(dirname `readlink -f $0`)
fi

SCRIPT_NAME=`basename "$0"`
# Source common_utils to use log_message.
. $current_dir/common_utils.sh

# (getopts call below must contain every possible options. Reference $RMAN_SHELL_SCRIPT_OPTIONS
#  to get the full list of options)
#
while getopts "${RMAN_SHELL_SCRIPT_OPTIONS}" option; do
  case $option in
    T)
      TAG=$OPTARG;;
    I)
      ID=$OPTARG;;
    s)
      is_physical_standby=1;;
    A)
      archivelog_keep_days=$OPTARG;;
    r)
      RMAN_BACKUP_TYPE=$OPTARG;;
    b)
      BACKUP_LEVEL=$OPTARG;;
    c)
      CROSSCHECK=1;;
    O)
      SBT_LIB_PATH=$OPTARG;;
    V)
      VIPS_LIST=$OPTARG;;
    W)
      VIEW_PATH=$OPTARG;;
    M)
      mount_point_list=$OPTARG;;
    m)
      SECTION_SIZE_CMD="SECTION SIZE $OPTARG";;
    f)
      FILESPERSET_CMD="FILESPERSET=$OPTARG";;
    G)
      ARCHIVELOG_BASE_TIME=$OPTARG;;
    *) ;;
  esac
done

# Modify how to invoke Oracle binaries if the agent is installed with 'root'
setup_invoke_oracle_binaries

#
# Process above options
#
SWITCH_LOG_COMMAND=''
[[ -z $is_physical_standby ]] && SWITCH_LOG_COMMAND="sql 'alter system archive log current';"

if [ ! -z "${RMAN_BACKUP_TYPE}" ] && [ "${RMAN_BACKUP_TYPE}" = "${SBT_BACKUP}" ]; then
  DEVICE_TYPE="SBT"
else
  DEVICE_TYPE="disk"
fi

ARCHIVELOG_DELETE_COMMAND=''
[[ ! -z $archivelog_keep_days ]] && ARCHIVELOG_DELETE_COMMAND="delete noprompt archivelog until time 'SYSDATE-${archivelog_keep_days}' backed up 1 times to device type ${DEVICE_TYPE};"

if [ "${RMAN_BACKUP_TYPE}" != "${SBT_BACKUP}" ]; then
  # Parse the mount point list so that we can store pfile.
  IFS=',' read -r -a mount_array <<< "$mount_point_list"
  if [[ ${#mount_array[@]} -eq 0 ]]; then
    exit_handler 1 "##### No mount path provided."
  fi
  if [[ "${mount_array[0]}" == *"/" ]]; then
    WHITELISTED_PATH_LIST+=( "${mount_array[0]}.*" )
  else
    WHITELISTED_PATH_LIST+=( "${mount_array[0]}/.*" )
  fi

  # Path in cohesity view where we store newly created pfile
  PFILE_PATH="${mount_array[0]}/pfile_backup.ora"

  if [ -f "$PFILE_PATH" ]; then
    log_message "deleting pfile from old backup run."
    safe_rm "$PFILE_PATH";
  fi

  log_message "Creating a new pfile from memory and saving it in cohesity view";

  timeout_cmd=`which timeout`
  if [ $? -ne 0 ]; then
    log_message "Timeout command not added to path. Not using."
    timeout_cmd=""
  else
    timeout_cmd="${timeout_cmd} 60"
  fi

  # If db_recovery_file_dest is out of space we are stuck in creating pfile and
  # don't progress. Hence putting a timeout of 60 second.
  ${timeout_cmd} $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF 2>&1
  CREATE PFILE='$PFILE_PATH' FROM MEMORY;
EOF

  if [ $? -eq 124 ]; then
    log_message "##### Creating pfile from memory timed out."
  fi

  if [ -f "$PFILE_PATH" ]; then
    log_message "Successfully Created pfile from memory"
  else
    log_message "##### Error creating pfile from memory"
  fi
fi

SPFILE_BACKUP_COMMAND=''
output=`$SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF 2>&1
   show parameter spfile;
EOF`
spfile=`echo $output | cut -d " " -f 9`

[[ ! -z $spfile ]] && SPFILE_BACKUP_COMMAND="backup spfile tag '$TAG';"

# Setting backup command based on type(imagecopy, backupsets).
DB_BACKUP_COMMAND="backup ${SECTION_SIZE_CMD} ${FILESPERSET_CMD} incremental level 1 for recover of copy with tag '${TAG}' database;"
DATAFILE_MERGE_COMMAND="RECOVER COPY OF DATABASE WITH TAG '$TAG';"
if [ "${RMAN_BACKUP_TYPE}" != ${IMAGECOPY_BACKUP} ]; then
  DB_BACKUP_COMMAND="backup ${SECTION_SIZE_CMD} ${FILESPERSET_CMD} incremental level ${BACKUP_LEVEL} tag '${TAG}' database;"
  DATAFILE_MERGE_COMMAND=""
fi

if [[ -z $ARCHIVELOG_BASE_TIME ]]; then
  LOG_BACKUP_COMMAND="backup tag '${TAG}' archivelog all not backed up 1 times;"
else
  get_local_time LOCAL_ARCHIVELOG_BASE_TIME $ARCHIVELOG_BASE_TIME "$connect_string"
  LOG_BACKUP_COMMAND="backup tag '${TAG}' archivelog from time '${LOCAL_ARCHIVELOG_BASE_TIME}';"
  LOG_BACKUP_COMMAND="$LOG_BACKUP_COMMAND backup tag '${TAG}' archivelog until time '${LOCAL_ARCHIVELOG_BASE_TIME}' not backed up 1 times;"
fi

CROSSCHECK_CMD=''

# Crosscheck backups and datafilecopies to improvide the clone performance.
if [[ ! -z $CROSSCHECK ]]; then
  CROSSCHECK_CMD="CROSSCHECK BACKUP OF DATABASE TAG '$TAG'; DELETE NOPROMPT EXPIRED BACKUP OF DATABASE TAG '$TAG';"
  CROSSCHECK_CMD="${CROSSCHECK_CMD} CROSSCHECK BACKUP OF CONTROLFILE TAG '$TAG'; DELETE NOPROMPT EXPIRED BACKUP OF CONTROLFILE TAG '$TAG';"
  CROSSCHECK_CMD="${CROSSCHECK_CMD} CROSSCHECK BACKUP OF SPFILE TAG '$TAG'; DELETE NOPROMPT EXPIRED BACKUP OF SPFILE TAG '$TAG';"
  CROSSCHECK_CMD=${CROSSCHECK_CMD}" crosscheck datafilecopy tag '$TAG';delete NOPROMPT expired datafilecopy tag '$TAG';"

  $RMAN_PATH msglog=${LOG_FILE} APPEND <<EOF
    connect target '${connect_string}${rman_privilege}';
    set echo on;
    set command id to '$ID';
    # Use rule optimizer for crosscheck - Oracle Doc ID 339964.1
    sql 'alter session set optimizer_mode=RULE';
    $CROSSCHECK_CMD
EOF
fi

# Setting device type base on rman backup(imagecopy, sbt).
sbt_channel_vec="";
sbt_channel_clear_vec="";

if [ "${RMAN_BACKUP_TYPE}" = "${SBT_BACKUP}" ]; then
  # Parse the mount point list.
  IFS=',' read -r -a vips_array <<< "$VIPS_LIST"
  for index in "${!vips_array[@]}"
  do
    sbt_channel_vec="${sbt_channel_vec} allocate channel c$((index+1)) device type sbt PARMS 'SBT_LIBRARY=${SBT_LIB_PATH}, SBT_PARMS=(mount_path=${VIEW_PATH}, vips=${vips_array[$index]}, gflag-name=sbt_use_grpc, gflag-value=true)';";
    sbt_channel_clear_vec="${sbt_channel_clear_vec} RELEASE CHANNEL c$((index+1));";
  done

  log_message "============================"
  log_message $sbt_channel_vec
  log_message $sbt_channel_clear_vec
  log_message "============================"
fi

log_message "Local channels to be allocated: ${channel_vec}";
log_message "Local channels to be released: ${channel_clear_vec}";
log_message "Remote channels to be released: ${remote_channel_clear_vec}";

#
# Execute the RMAN command to backup the database.
# Delete source archived logs according to input option.
#
$RMAN_PATH msglog=${LOG_FILE} APPEND <<EOF
    connect target '${connect_string}${rman_privilege}';
    set echo on;
    set command id to '$ID';
    sql 'alter session set NLS_DATE_FORMAT = "DD:MM:YYYY-HH24:MI:SS"';
    CONFIGURE CONTROLFILE AUTOBACKUP OFF;
    # Use rule optimizer for deleting archive logs - Oracle Doc ID 1589307.1
    sql 'alter session set optimizer_mode=RULE';
    run {
      $channel_vec
      $remote_channel_vec
      $sbt_channel_vec
      $DB_BACKUP_COMMAND
      backup current controlfile tag '$TAG';
      $SWITCH_LOG_COMMAND
      $LOG_BACKUP_COMMAND
      # spfile backup must be the last backupset. Recovery logic looks for
      # the highest backup set that is not of controlfile type, and does not
      # have archived logs in it, to identify this as the spfile backupset.
      # Incremental datafile backupset has the same characteristic, except
      # we only look for the last backup set that has this characteristic as
      # the spfile.
      $SPFILE_BACKUP_COMMAND
      $DATAFILE_MERGE_COMMAND
      $ARCHIVELOG_DELETE_COMMAND
      $sbt_channel_clear_vec
      $channel_clear_vec
      $remote_channel_clear_vec
    }
    set echo off;
    quit;
EOF
