#!/bin/bash
# Copyright 2020 Cohesity Inc.
#
# Author: Manipal Singh
#
# This is a wrapper script to migrate db running on Cohesity view to
# primary storage on oracle host.
# Workflow:
# 1. Find and move each datafile that is there on cohesity view to the provided
#    location. Datafile that are not on cohesity view will not be migrated.
# 2. Fetch redo log info and migrate
#   a) Get redo logs infos containing group, status, size and members.
#   b) Create same number of logs files with same size on new destination.
#   c) Switch log files to make old ones INACTIVE
#   d) Once an old redo is INACTIVE we drop it.
#   NOTE: If some members are on cohesity, rest not, then all of them will be
#   migrated to new location.
#
# It also appends the output at /tmp/cohesity_migrate_db_{task_id}.INFO file


# Connect string is passsed as stdin.
read connect_string;

LOG_FILE=""
SQL_OUTPUT_DELIM="#"
# Maximum consecutive attempts for dropping redo logs by switching redologs.
MAX_SWITCHOVER_ATTEMPTS=10

# Which OS Cohesity agent is running on.
OS=$(uname -a | cut -d " " -f 1)

if [ ${OS} = "AIX" ]; then
  # TODO(aabhas): find replacement for readlink on AIX.
  current_dir=$(dirname $0)
else
  current_dir=$(dirname `readlink -f $0`)
fi

SCRIPT_NAME=`basename "$0"`
# Source needed library functions.
. $current_dir/common_utils.sh

log_message() {
  msg=$*
  TS=`date "+I%m%d %H:%M:%S"`
  if [ "$LOG_FILE" == "" ]; then
    echo "$TS [$SCRIPT_NAME] `basename ${MOUNTPOINT_PATTERN}`: ${msg}"
  else
    echo "$TS [$SCRIPT_NAME] `basename ${MOUNTPOINT_PATTERN}`: ${msg}" >> ${LOG_FILE}
  fi
}

# Cleanup function.
# Logic: Cleanup function will be called on any failure/error or any interrupt
# signals. Once the cleanup is hit, we will check if sql is running and try to
# stop it.
cleanup () {
  result=$?
  log_message "Trap Line number: $1"
  log_message "Exit status is $result";

  # Check if SQL is running, possibly during cancellation of jobs.
  # Stop the SQL if true.
  if [[ $SQL_PID -ne 1 ]] && [ ps -p $SQL_PID > /dev/null 2>&1 ]; then
    log_message "Stopping sql process."
    kill -INT $SQL_PID
    CUR_SQL_PID=$SQL_PID
    SQL_PID=0
    log_message "Waiting for sql process to be killed."
    wait $CUR_SQL_PID
  fi

  trap - EXIT
  exit $result
}

# Migrate datafile from given source path to destination path.
# Params:
# $1: Source path of the datafile.
# $2: Target path of the datafile.
# $3: Name of the container to which the datafile belongs to.
# $4: Container id where the container belongs.
migrate_datafile() {
  if [[ $# -ne 4 ]]; then
    log_message "Invalid number of arguments. Expected: 4, provided: $@"
    exit 1
  fi
  log_message "Migrating datafile on container $3 from $1 to $2"

  if [ $4 -ne 0 ]; then
    alter_container_cmd="ALTER SESSION SET CONTAINER=$3;";
    log_message "alter container command: ${alter_container_cmd}"
  fi
  $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE} &
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    ${alter_container_cmd}
    ALTER DATABASE MOVE DATAFILE '$1' to '$2';
EOF

  # Hold the SQL process PID in a variable.
  # Logic, SQL runs in the background and we hold the PID.
  # Wait for the process to complete in the normal workflow.
  # If cancellation is invoked, then interrupt the sql process.
  SQL_PID=$!
  log_message "SQL process started with PID: $SQL_PID"

  # Wait for background SQL to complete.
  log_message "Waiting for datafile migration process to complete."
  wait $SQL_PID
  result=$?
  if [[ $result -ne 0 ]]; then
    log_message "Error in datafile: $datafile_path migration."
    exit 1;
  fi
  SQL_PID=0
}

# This will populate result in ALL_REDOLOGS_INFO_LIST.
# The result will be have ${SQL_OUTPUT_DELIM} separated 4 columns.
# The columns are Group number, Status, Max size of member and Total number of members in the group.
# Sample output:
# 1#CURRENT#20#2
# 2#INACTIVE#30#3
get_all_redologs_info_list() {
  local sql_cmd="$SQLPLUS_PATH -S $connect_string $sqlplus_privilege @$SQL_SCRIPT_PATH/$REDOLOG_INFO_SCRIPT
           $os_error $sql_error $ID $SQL_OUTPUT_DELIM $SQL_SCRIPT_PATH %"
  if [[ $? -ne 0 ]]; then
    log_message "Error in fetching all redolog list."
    exit 1;
  fi
  ALL_REDOLOGS_INFO_LIST=`$sql_cmd`
  ALL_REDOLOGS_INFO_LIST=`echo "$ALL_REDOLOGS_INFO_LIST" | xargs`
  IFS=' ' read -r -a ALL_REDOLOGS_INFO_LIST <<< "${ALL_REDOLOGS_INFO_LIST}"
}

# This will populate result in COHESITY_REDOLOGS_INFO_LIST.
# The result will have ${SQL_OUTPUT_DELIM} separated 4 columns.
# The columns are Group number, Status, Max size of member and Total number of members in the group.
get_cohesity_redologs_info_list() {
  local sql_cmd="$SQLPLUS_PATH -S $connect_string $sqlplus_privilege @$SQL_SCRIPT_PATH/$REDOLOG_INFO_SCRIPT
           $os_error $sql_error $ID $SQL_OUTPUT_DELIM $SQL_SCRIPT_PATH ${MOUNTPOINT_PATTERN}%"
  COHESITY_REDOLOGS_INFO_LIST=`$sql_cmd`
  if [[ $? -ne 0 ]]; then
    log_message "Error in fetching cohesity redolog list."
    exit 1;
  fi
  COHESITY_REDOLOGS_INFO_LIST=`echo "$COHESITY_REDOLOGS_INFO_LIST" | xargs`
  IFS=' ' read -r -a COHESITY_REDOLOGS_INFO_LIST <<< "${COHESITY_REDOLOGS_INFO_LIST}"
}

# This will populate the result in DATAFILE_INFO_LIST for the datafile that are
# on cohesity mountpoint.
# The result will have ${SQL_OUTPUT_DELIM} separated 3 columns.
# The columns are container id,container name and datafile path.
get_cohesity_datafile_info_list() {
  local sql_cmd="$SQLPLUS_PATH -S $connect_string $sqlplus_privilege @$SQL_SCRIPT_PATH/$DATAFILE_INFO_SCRIPT
            $os_error $sql_error $ID $SQL_OUTPUT_DELIM $SQL_SCRIPT_PATH ${MOUNTPOINT_PATTERN}%"
  log_message "Running datafile info cmd: $sql_cmd"
  DATAFILE_INFO_LIST=`$sql_cmd`
  if [[ $? -ne 0 ]]; then
    log_message "Error in fetching datafile information."
    exit 1;
  fi
}

# Helper method to switch redo logs
switch_redologs() {
  $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE}
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    alter system switch logfile;
    alter system checkpoint;
EOF
  if [[ $? -ne 0 ]]; then
    log_message "Failure during switching redologs."
    exit 1;
  fi
}

# Helper method to get the next usable redolog group number.
# Sets value in NEXT_REDOLOG_GROUP_NUM.
get_next_redolog_group_num() {
  NEXT_REDOLOG_GROUP_NUM=`$SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF 2>&1
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    set heading off;
    set verify off;
    set feedback off;
    set pagesize 5000;
    set linesize 30000;
    select max(group#) from V\\$LOG;
EOF`
  if [[ $? -ne 0 ]]; then
    log_message "Error in fetching next usable redolog group number."
    exit 1;
  fi
  if [[ $(($NEXT_REDOLOG_GROUP_NUM * 2)) -lt $MAX_SWITCHOVER_ATTEMPTS ]]; then
    MAX_SWITCHOVER_ATTEMPTS=$(($NEXT_REDOLOG_GROUP_NUM * 2))
  fi

  NEXT_REDOLOG_GROUP_NUM=$(($NEXT_REDOLOG_GROUP_NUM + 1 ))
}

# Helper method to get number of redolog groups to be added.
get_num_groups_to_add() {
  local total_redolog_groups_on_cohesity=${#COHESITY_REDOLOGS_INFO_LIST[@]}
  local total_redolog_groups_outside_cohesity=$(( ${#ALL_REDOLOGS_INFO_LIST[@]} - $total_redolog_groups_on_cohesity ))

  # We will add new groups only if groups outside of cohesity view are lesser than
  # that on cohesity view.
  NUM_GROUPS_TO_ADD=$(($total_redolog_groups_on_cohesity - $total_redolog_groups_outside_cohesity))
  log_message "total_redolog_groups_on_cohesity: $total_redolog_groups_on_cohesity"
  log_message "total_redolog_groups_outside_cohesity: $total_redolog_groups_outside_cohesity"
  if [[ $NUM_GROUPS_TO_ADD -lt 0 ]]; then
    # Number of view outside cohesity are more, no need to add.
    NUM_GROUPS_TO_ADD=0
  fi
}

usage()
{
  echo "Invoke the script with following arguments :-"
  echo -e "\t -A|--db-sid                       Oracle source database SID value."
  echo -e "\t -d|--datafile-info-script         Script to get datafile information."
  echo -e "\t -D|--redolog-destination          Redo log group destination."
  echo -e "\t -F|--log-file-path                Rman log file path."
  echo -e "\t -G|--redolog-info-script          Script to get redolog information."
  echo -e "\t -H|--oracle-home                  Oracle home directory."
  echo -e "\t -I|--id                           Operation ID."
  echo -e "\t -L|--migrate-destination-location Target location of database files."
  echo -e "\t -M|--mount-points                 CSV mount point list."
  echo -e "\t -O|--os-error                     OS error code."
  echo -e "\t -Q|--sql-error                    SQL error code."
  echo -e "\t -P|--sql-scripts-path             Path of Sqlplus scripts."
  echo -e "\t -R|--rman-path                    Path to RMAN binary."
  echo -e "\t -S|--sqlplus-path                 Path to source SQLPLUS binary."
}

##### Main

while getopts "A:d:D:F:G:H:I:L:M:O:Q:P:R:S:" option; do
  case $option in
    A)
      DB_SID=$OPTARG;;
    d)
      DATAFILE_INFO_SCRIPT=$OPTARG;;
    D)
      REDOLOG_DEST=$OPTARG;;
    F)
      LOG_FILE_PATH=$OPTARG;;
    G)
      REDOLOG_INFO_SCRIPT=$OPTARG;;
    H)
      ORACLE_HOME_DIR=$OPTARG;;
    I)
      ID=$OPTARG;;
    L)
      DEST_LOCATION=$OPTARG;;
    M)
      MOUNTPOINT_PATTERN=$OPTARG;;
    O)
      os_error=$OPTARG;;
    Q)
      sql_error=$OPTARG;;
    P)
      SQL_SCRIPT_PATH=$OPTARG;;
    R)
      RMAN_PATH=$OPTARG;;
    S)
      SQLPLUS_PATH=$OPTARG;;
    *) echo "Invalid Option passed"; usage; exit 1 ;;
  esac
done

export ORACLE_SID=$DB_SID
export ORACLE_HOME=$ORACLE_HOME_DIR
export ORACLE_BASE=$($ORACLE_HOME/bin/orabase)
export NLS_DATE_FORMAT='DD:MM:YYYY-HH24:MI:SS'

if [[ "x$LOG_FILE_PATH" == "x" ]]; then
  LOG_FILE_PATH="/tmp"
fi
LOG_FILE=${LOG_FILE_PATH}/migrate_db_${ID}.INFO
# Update rman log file permissions.
touch $LOG_FILE
sudo chmod 766 $LOG_FILE

trap 'cleanup $LINENO' EXIT INT TERM

if [ -z "$RMAN_PATH" ] || [ -z "$SQLPLUS_PATH" ] || [ -z "$DB_SID" ]; then
  log_message "Script parameters missing."
  log_message "#####$RMAN_PATH, $SQLPLUS_PATH, $DB_SID";
  usage;
  exit 1;
fi

if [ -z "$connect_string" ]; then
  log_message "Auth string is missing"
  usage;
  exit 1;
fi

if [ -z "$DEST_LOCATION" ] || [ -z "$MOUNTPOINT_PATTERN" ]; then
  log_message "Script location parameters missing.";
  log_message "######$DEST_LOCATION, $MOUNTPOINT_PATTERN";
  usage;
  exit 1;
fi

if [ -z "$REDOLOG_DEST" ]; then
  log_message "Redolog destination missing.";
  usage;
  exit 1;
fi

# NOTE: We are assuming that either both path will be ASM or none.
f_check_is_asm_path $DEST_LOCATION
f_check_is_asm_path $REDOLOG_DEST

# For non ASM path check if directory exists.
if [[ "$IS_ASM" == "false" ]] && [ ! -d "$DEST_LOCATION" ]; then
  log_message "Destination location: ${DEST_LOCATION} does not exists.";
  usage;
  exit 1;
fi

# For non ASM path check if directory exists.
if [[ "$IS_ASM" == "false" ]] && [ ! -d "$REDOLOG_DEST" ]; then
  log_message "Redo log destination location: ${REDOLOG_DEST} does not exists.";
  usage;
  exit 1;
fi

if [ -z "$REDOLOG_INFO_SCRIPT" ] || [ -z "$DATAFILE_INFO_SCRIPT" ]; then
  log_message "Script parameter missing."
  usage;
  exit 1;
fi

# Update the MOUNTPOINT_PATTERN to actual normalized path.
log_message "Updating mountpoint path: ${MOUNTPOINT_PATTERN}"
MOUNTPOINT_PATTERN=`readlink -m ${MOUNTPOINT_PATTERN}`
log_message "Mountpoint path updated to: ${MOUNTPOINT_PATTERN}"

# Modify how to invoke Oracle binaries if the agent is installed with 'root'
setup_invoke_oracle_binaries

log_message "Shell Environment:"
log_message "`env`"

f_set_privilege "$ORACLE_HOME_DIR" "$connect_string"

# Modify the db file creation destination.
log_message "Modifying DB create and recovery file dest to ${DEST_LOCATION}"
$SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE}
  WHENEVER OSERROR EXIT $os_error
  WHENEVER SQLERROR EXIT $sql_error
  ALTER SYSTEM SET DB_CREATE_FILE_DEST='${DEST_LOCATION}' SCOPE=BOTH;
  ALTER SYSTEM SET DB_RECOVERY_FILE_DEST='${DEST_LOCATION}' SCOPE=BOTH;
EOF
if [[ $? -ne 0 ]]; then
  log_message "Error in modifying DB Create/recovery destination to: ${DEST_LOCATION}."
  exit 1;
fi

echo "--------------------------------------------------------------------" >> ${LOG_FILE}
log_message "Starting datafile migration."

get_cohesity_datafile_info_list

for datafile_info in ${DATAFILE_INFO_LIST}
do
  IFS=${SQL_OUTPUT_DELIM} read -r -a datafile_info_array <<< "${datafile_info}"
  if [[ ${#datafile_info_array[@]} -ne 3 ]]; then
    # We will skip any invalid datafile logs row, we will check for datafile on
    # Cohesity view in post migration prechecks.
    log_message "Invalid datafile info row: ${datafile_info}"
    continue
  fi

  cont_id=${datafile_info_array[0]}
  cont_name=${datafile_info_array[1]}
  datafile_path=${datafile_info_array[2]}

  datafile_name=$(basename $datafile_path)
  log_message "Migrating datafile $datafile_path with container name $cont_name"
  if [[ "$IS_ASM" == "false" ]]; then
    migrate_datafile ${datafile_path} ${DEST_LOCATION}/${datafile_name} ${cont_name} ${cont_id}
  else
    migrate_datafile ${datafile_path} ${DEST_LOCATION} ${cont_name} ${cont_id}
  fi
done

log_message "Datafile migration completed."
echo "--------------------------------------------------------------------" >> ${LOG_FILE}
log_message "Starting addition of redo logs."

# Get the smallest group number that can be used for creating new redolog groups.
get_next_redolog_group_num
log_message "Next usable redolog group number: ${NEXT_REDOLOG_GROUP_NUM}"

get_all_redologs_info_list
get_cohesity_redologs_info_list

get_num_groups_to_add
log_message "Number of redolog groups to add: ${NUM_GROUPS_TO_ADD}"

# Prepare the add redologs group command
redolog_group_add_cmd=""
for ((ii=0;ii<${NUM_GROUPS_TO_ADD};ii++))
do
  redolog_info=${COHESITY_REDOLOGS_INFO_LIST[$ii]}
  IFS=${SQL_OUTPUT_DELIM} read -r -a redolog_info_array <<< "${redolog_info}"
  if [[ ${#redolog_info_array[@]} -ne 4 ]]; then
    # We will skip any invalid redo logs row, we will check for redo groups
    # on cohesity view after migration is completed.
    log_message "Invalid redo log info row: ${redologs_info}"
    continue
  fi

  group_num=${redolog_info_array[0]}

  # Formulate member path based on destination location and number of members.
  member_path=""
  for ((jj=0;jj<${redolog_info_array[3]};jj++))
  do
    new_member_path=""
    if [[ "$IS_ASM" == "false" ]]; then
      new_member_path="/redo_${group_num}_${jj}.log"
    fi

    if [[ "x$member_path" == "x" ]]; then
      member_path="'${REDOLOG_DEST}${new_member_path}'"
    else
      member_path="${member_path},'${REDOLOG_DEST}${new_member_path}'"
    fi
  done

  member_size=${redolog_info_array[2]}
  log_message "Adding group: ${NEXT_REDOLOG_GROUP_NUM} correponding to existing group: ${group_num}"

  redolog_group_add_cmd="${redolog_group_add_cmd} ALTER DATABASE ADD LOGFILE GROUP ${NEXT_REDOLOG_GROUP_NUM} (${member_path}) size ${member_size}m;\n"
  NEXT_REDOLOG_GROUP_NUM=$(($NEXT_REDOLOG_GROUP_NUM + 1))
done

if [[ "x$redolog_group_add_cmd" != "x" ]]; then
  redolog_group_add_cmd=`echo -e ${redolog_group_add_cmd}`
  log_message "Adding redo logs, running SQL command: ${redolog_group_add_cmd}"

  $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE} &
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    ${redolog_group_add_cmd}
EOF
  # Hold the SQL process PID in a variable.
  SQL_PID=$!
  log_message "SQL process started with PID: $SQL_PID"

  # Wait for background SQL to complete.
  log_message "Waiting for redo logs addition process to complete."
  wait $SQL_PID
  result=$?
  SQL_PID=0
  if [[ $result -ne 0 ]]; then
    log_message "Failure while adding redo logs."
    exit 1
  fi
fi

log_message "Redo log groups added."
echo "--------------------------------------------------------------------" >> ${LOG_FILE}

# In case of RAC to standalone clone creation there might be threads pointing
# to Unknown instances. If these threads are not disable them some redo logs
# might remain in UNUSED state and we won't be able to drop them.
MAX_THREAD_NUMBER=`$SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF 2>&1
  WHENEVER OSERROR EXIT $os_error
  WHENEVER SQLERROR EXIT $sql_error
  set heading off;
  set verify off;
  select max(thread#) from V\\$LOG;
EOF`
if [[ $? -ne 0 ]]; then
  log_message "Error in getting max thread number."
  exit 1;
fi
if [[ $MAX_THREAD_NUMBER -ge 2 ]]; then
  log_message "Disabling threads from 2 till ${MAX_THREAD_NUMBER}."
  for ((ii=2;ii<=${MAX_THREAD_NUMBER};ii++))
  do
    thread_disable_cmd="${thread_disable_cmd} alter database disable thread ${ii};"
  done
  log_message "Executing command: ${thread_disable_cmd}"
  $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE}
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    ${thread_disable_cmd}
EOF
  if [[ $? -ne 0 ]]; then
    log_message "Error in disabling thread."
    exit 1;
  fi
fi

log_message "Starting deletion of old redo log groups."
log_message "Redolog groups to be dropped: ${groups_to_be_dropped[@]}"

switchover_attempts=0
while [[ ${#COHESITY_REDOLOGS_INFO_LIST[@]} -gt 0 ]]; do
  if [[ $switchover_attempts -ge $MAX_SWITCHOVER_ATTEMPTS ]]; then
    log_message "Maximum consecutive attempts for switchover exceeded, unable to remove old redolog groups."
    exit 1;
  fi

  # Switch logfiles to get old redologs into dropped state.
  switch_redologs
  # Get latest redolog information.
  get_cohesity_redologs_info_list
  redolog_drop_cmd=""
  for redolog_info in ${COHESITY_REDOLOGS_INFO_LIST}
  do
    IFS=${SQL_OUTPUT_DELIM} read -r -a redolog_info_array <<< "${redolog_info}"
    # Check if the redolog group can be dropped, i.e status is INACTIVE or UNUSED.
    if [[ "${redolog_info_array[1]}" != "UNUSED" ]] && [[ "${redolog_info_array[1]}" != "INACTIVE" ]]; then
      continue
    fi

    redolog_drop_cmd="${redolog_drop_cmd}alter database drop logfile group ${redolog_info_array[0]};\n"
    # Reset switchover attempts after we find atleast one droppable group.
    switchover_attempts=0
  done

  if [[ "x$redolog_drop_cmd" == "x" ]]; then
    # There might be cases when after switchover we will not be able to
    # find any droppable redologs.
    switchover_attempts=$(($switchover_attempts + 1))
    continue
  fi

  redolog_drop_cmd=`echo -e ${redolog_drop_cmd}`
  log_message "Dropping redo logs, running SQL command: ${redolog_drop_cmd}"
  $SQLPLUS_PATH -S $connect_string $sqlplus_privilege << EOF >> ${LOG_FILE} &
    WHENEVER OSERROR EXIT $os_error
    WHENEVER SQLERROR EXIT $sql_error
    ${redolog_drop_cmd}
EOF
  # Hold the SQL process PID in a variable.
  SQL_PID=$!
  log_message "SQL process started with PID: $SQL_PID"

  # Wait for background SQL to complete.
  log_message "Waiting for redo logs deletion process to complete."
  wait $SQL_PID
  result=$?
  SQL_PID=0
  if [[ $result -ne 0 ]]; then
    log_message "Failure while removing redo logs."
    exit 1
  fi
done

log_message "Deletion of old redo log groups completed."
echo "--------------------------------------------------------------------" >> ${LOG_FILE}
