#!/bin/bash
# Copyright 2019 Cohesity Inc.
#
# Author: Sandeep Tandekar
#
# This script performs the restore of database control file.
# It takes command ID, control file path as input and database id.
# Example :
# /home/oracle/cohesityagent/software/crux/bin/oracle_scripts//restore_control_file.sh
# -R /u01/app/oracle/product/11.2.0.3/db_1/bin/rman -H /u01/app/oracle/product/11.2.0.3/db_1
# -S v1 -C / -I cohesity_5704 -F /tmp -B 3312305298 -T True
# -A /home/oracle/cohesityagent/data/agent/tmp/nfs_oracle_mounts/oracle_5704_5_path0/5700_33toh3m8_1_1

# Read the stdin as first step.
read connect_string;

# Exit from the script if any of the step fails.
set -e

LOG_FILE=""

# Which OS Cohesity agent is running on.
OS=$(uname -a | cut -d " " -f 1)

if [ ${OS} = "AIX" ]; then
  # TODO(aabhas): find replacement for readlink on AIX.
  current_dir=$(dirname $0)
else
  current_dir=$(dirname `readlink -f $0`)
fi

SCRIPT_NAME=`basename "$0"`
# Source needed library functions.
. $current_dir/common_utils.sh


usage()
{
  echo "Invoke the script with following arguments :-"
  echo -e "\t -R|--rman-path        Path to RMAN binary."
  echo -e "\t -H|--oracle-home      Path to ORACLE home directory."
  echo -e "\t -S|--oracle-sid       Oracle database SID value."
  echo -e "\t -T|--stop-db          If set then stop the dababase."
  echo -e "\t -M|--start-mount-db   If set then start the database in mount mode after control file restore."
  echo -e "\t -I|--id               Operation ID."
  echo -e "\t -F|--log-file-path    Rman log file path."
  echo -e "\t -A|--control-file     Database control file path."
  echo -e "\t -B|--db-id            Database Id."
  echo -e "\t -s|--is-standby       It is set if the DB is working in standby role."
  echo -e "\t -X|--rman-backup-type Type of rman backup(ImageCopy or BackupSets)."
  echo -e "\t -O|--sbt-lib-path     SBT library path on this Oracle host."
  echo -e "\t -V|--vips-list        CSV VIPs list."
  echo -e "\t -W|--full-view-path   Full Cohesity view path where control file is available."
  echo -e "\t -Q|--sqlplus-path     Path to source SQLPLUS binary."
  echo -e "\t -f|--parameter-file-loc Location where spfile/pfile needs to be restored"
  echo -e "\t -U|--db-unique-name   Database Unique Name"
  echo -e "\t -p|--restore-spfile   whether to restore spfile or not."
}

##### Main


while getopts "R:H:S:T:M:I:F:A:B:sX:O:V:W:Q:f:U:p" option; do
  case $option in
    R)
      RMAN_PATH=$OPTARG;;
    H)
      export ORACLE_HOME=$OPTARG;;
    S)
      export ORACLE_SID=$OPTARG;;
    T)
      STOP_DB=$OPTARG;;
    M)
      START_MOUNT_DB=$OPTARG;;
    I)
      ID=$OPTARG;;
    F)
      LOG_FILE_PATH=$OPTARG;;
    A)
      CONTROL_FILE=$OPTARG;;
    B)
      DATABASE_ID=$OPTARG;;
    s)
      IS_STANDBY=1;;
    X)
      RMAN_BACKUP_TYPE=$OPTARG;;
    O)
      SBT_LIB_PATH=$OPTARG;;
    V)
      VIPS_LIST=$OPTARG;;
    W)
      VIEW_PATH=$OPTARG;;
    Q)
      SQLPLUS_PATH=$OPTARG;;
    f)
      PARAMETER_FILE_LOC=$OPTARG;;
    U)
      DB_UNIQUE_NAME=$OPTARG;;
    p)
      RESTORE_SPFILE=1;;
    *) echo "Invalid Option passed"; usage; exit 1 ;;
  esac
done

if [ -z "${RMAN_PATH}" ] || [ -z "${ORACLE_HOME}" ] || [ -z "${ORACLE_SID}" ] ||  [ -z "${connect_string}" ]; then
  echo "RMAN configuration parameters or authentication is missing."
  echo "*** ${RMAN_PATH}, ${ORACLE_HOME}, ${ORACLE_SID}";
  usage;
  exit 1;
fi

if [ -z "${CONTROL_FILE}" ] || [ -z "${ID}" ] || [ -z "${DATABASE_ID}" ]; then
  echo "Database parameters missing."
  echo "*** ${CONTROL_FILE}, ${ID}, ${DATABASE_ID}";
  usage;
  exit 1;
fi

# Checking SBT params.
if [ "${RMAN_BACKUP_TYPE}" = "${SBT_BACKUP}" ]; then
  if [ -z "$VIEW_PATH" ] || [ -z "$SBT_LIB_PATH" ] || [ -z "$VIPS_LIST" ]; then
    echo "SBT configuration parameters missing."
    echo "*** $VIEW_PATH, $SBT_LIB_PATH, $VIPS_LIST";
    usage;
    exit 1;
 fi
fi

# Export PATH and other enviroment params.
export ORACLE_BASE=$($ORACLE_HOME/bin/orabase)
export NLS_DATE_FORMAT='DD:MM:YYYY-HH24:MI:SS'
export LD_LIBRARY_PATH="$LD_LIBRARY_PATH:${DEST_ORACLE_HOME}/lib"
export PATH="$PATH:${DEST_ORACLE_HOME}/bin"

if [[ "x$LOG_FILE_PATH" == "x" ]]; then
  LOG_FILE_PATH="/tmp"
fi
LOG_FILE=${LOG_FILE_PATH}/rman_shell_${ID}.INFO

echo "-------------------------" >> ${LOG_FILE}
log_message "Date: " `date`
log_message "Wrapper script args: $@"
log_message "Shell environment:"
log_message "`env`"

# Modify how to invoke Oracle binaries if the agent is installed with 'root'
setup_invoke_oracle_binaries

f_set_privilege "$ORACLE_HOME" "$connect_string"

# After this point we want to handle the error.
set +e


log_message "Shutting down the database before control file restore."
if [ ! -z "${STOP_DB}" ]; then
  ${RMAN_PATH} >>${LOG_FILE} <<EOF
  connect target '${connect_string}${rman_privilege}';
  set command id to '${ID}';
  SHUTDOWN ABORT;
EOF

  log_message "Sleeping for 10 seconds."
  sleep 10;
fi

# Setting device type base on rman backup(imagecopy, sbt).
sbt_channel_vec="";
sbt_clear_channel_vec="";
if [ "${RMAN_BACKUP_TYPE}" = "${SBT_BACKUP}" ]; then
  # Parse the mount point list.
  IFS=',' read -r -a vips_array <<< "$VIPS_LIST"
  for index in "${!vips_array[@]}"
  do
    sbt_channel_vec="${sbt_channel_vec} allocate channel c$((index+1)) device type sbt PARMS 'SBT_LIBRARY=${SBT_LIB_PATH}, SBT_PARMS=(mount_path=${VIEW_PATH}, vips=${vips_array[$index]}, gflag-name=sbt_use_grpc, gflag-value=true)';";
    sbt_clear_channel_vec="${sbt_clear_channel_vec} RELEASE CHANNEL c$((index+1));";
  done

  log_message "============================"
  log_message $sbt_channel_vec
  log_message $sbt_clear_channel_vec
  log_message "============================"
fi

get_spfile_path() {
  # find mount path
  COHESITY_VIEW=$(dirname `readlink -f $CONTROL_FILE`)
  SPFILE_PREFIX="spfile_"
  SPFILE_PATH=`find ${COHESITY_VIEW} -name "*${SPFILE_PREFIX}*" | head -n 1`
}

validate_spfile_params() {
  if [ -z ${SPFILE_PATH} ] ; then
    log_message "Spfile path is not passed so skipping restoring spfile. This could happen if db was started with pfile."
    exit 1
  fi
  if [ ! -f ${SPFILE_PATH} ] ; then
    log_message "${SPFILE_PATH} does not exist on restore view, restore control file might fail."
    exit 1
  fi
}

restore_spfile_on_asm() {
  if [ -z ${DB_UNIQUE_NAME} ] ; then
    log_message "DB_UNIQUE_NAME is not passed so skipping restoring spfile on asm."
    return
  fi

  SRVCTL_SCRIPT_PATH="${current_dir}/srvctl_from_oracle_home.sh"
  if [ ! -f ${SRVCTL_SCRIPT_PATH} ]; then
    log_message "srvctl script path: ${SRVCTL_SCRIPT_PATH} does not exist, cannot restore spfile."
    exit 1
  fi
 log_message "Found spfile in restore view, trying to restore it"
  timestamp=$(date +%s)
  SPFILE_RESTORE_LOC="${PARAMETER_FILE_LOC}/spfile${DB_UNIQUE_NAME}.${timestamp}"
  SPFILE_CLAUSE="RESTORE SPFILE to '$SPFILE_RESTORE_LOC' from '$SPFILE_PATH';";
  ${RMAN_PATH} >>${LOG_FILE} <<EOF
    connect target '${connect_string}${rman_privilege}';
    SET ECHO ON;
    STARTUP NOMOUNT;
    ${SPFILE_CLAUSE}
EOF
  result=$?
  if [ ${result} -eq 0 ]; then
    log_message "Succesfully restored SPFILE on asm. Updating new spfile path on srvctl."
    ${SRVCTL_SCRIPT_PATH} $ORACLE_HOME bin/srvctl modify database -d ${DB_UNIQUE_NAME} -p ${SPFILE_RESTORE_LOC}
    result=$?
    if [ ${result} -eq 0 ]; then
      log_message "Succesfully updated spfile location on srvctl. Stopping the db";
      ${SRVCTL_SCRIPT_PATH} $ORACLE_HOME bin/srvctl stop database -d ${DB_UNIQUE_NAME}
      result=$?
      if [ ${result} -ne 0 ]; then
        log_message "Failed to stop db using srvctl after spfile restore."
        exit 1
      fi
      log_message "Starting database in NOMOUNT mode.";
      # Start all nodes of RAC in nomount state.
      ${SRVCTL_SCRIPT_PATH} $ORACLE_HOME bin/srvctl start database -d ${DB_UNIQUE_NAME} -o NOMOUNT
      log_message "Checking if all nodes are in NOMOUNT mode";
      ${SRVCTL_SCRIPT_PATH} $ORACLE_HOME bin/srvctl status database -d ${DB_UNIQUE_NAME} | grep "is not running"
      result=$?
      if [ ${result} -eq 0 ]; then
        log_message "Failed to start all nodes in nomount mode."
        exit 1
      fi
      log_message "Succesfully started all nodes in nomount mode."
      return
    fi
  fi
  log_message "spfile restore or updation failed, control file restore might fail because of it."
  exit 1
}

restore_spfile() {
  if [ "x$PARAMETER_FILE_LOC" == "x"  ]; then
    if [ ! -d "${ORACLE_HOME}/dbs" ]; then
      log_message "${ORACLE_HOME}/dbs folder does not exist and PARAMETER_FILE_LOC is also not passed. Cannot restore spfile/pfile."
      exit 1
    fi
  fi
  log_message "Found spfile in restore view, trying to restore it"
  SPFILE_CLAUSE="RESTORE SPFILE FROM '$SPFILE_PATH';"
  if [ "x$PARAMETER_FILE_LOC" == "x"  ]; then
    # Till 19c spfile and pfile are stored at ${ORACLE_HOME}/dbs folder
    # But from 21c they could be stored either at ${ORACLE_BASE}/dbs
    # or ${ORACLE_HOME}/dbs folder depending on whether ${ORACLE_HOME} is read-only.
    PARAMETER_FILE_LOC="${ORACLE_HOME}/dbs"
  fi
  SPFILE_RESTORE_LOC="${PARAMETER_FILE_LOC}/spfile${ORACLE_SID}.ora"
  SPFILE_CLAUSE="RESTORE SPFILE to '$SPFILE_RESTORE_LOC' from '$SPFILE_PATH';";
  # move existing spfile if any, else restore spfile will fail.
  if [ -f ${SPFILE_RESTORE_LOC} ] ; then
    timestamp=$(date +%s)
    BACKUP_SPFILE_LOC="$PARAMETER_FILE_LOC/spfile${ORACLE_SID}_${timestamp}.ora"
    log_message "spfile already exist at $SPFILE_RESTORE_LOC, moving old spfile to $BACKUP_SPFILE_LOC.";
    mv $SPFILE_RESTORE_LOC $BACKUP_SPFILE_LOC
  fi
  ${RMAN_PATH} >>${LOG_FILE} <<EOF
    connect target '${connect_string}${rman_privilege}';
    SET ECHO ON;
    STARTUP NOMOUNT;
    ${SPFILE_CLAUSE}
    SHUTDOWN ABORT;
EOF
  result=$?
  if [ ${result} -eq 0 ]; then
    log_message "Succesfully restored SPFILE."
    return
  fi
  log_message "Could not restore SPFILE."
  exit 1
}

if [ ! -z ${RESTORE_SPFILE} ]; then
  get_spfile_path
  f_check_is_asm_path $PARAMETER_FILE_LOC
  validate_spfile_params
  if [ "$IS_ASM" == "false" ]; then
    log_message "Trying to restore spfile/pfile on non-asm";
    restore_spfile
  else
    log_message "Trying to restore spfile on asm"
    restore_spfile_on_asm
  fi
else
  log_message "Not restoring spfile as part of controlfile restore."
fi

if [ ! -z ${IS_STANDBY} ]; then
  RESTORE_CLAUSE="RESTORE STANDBY CONTROLFILE FROM '${CONTROL_FILE}';"
  log_message "Restoring the standby control file ${CONTROL_FILE} for database sid ${ORACLE_SID}."
else
  RESTORE_CLAUSE="RESTORE PRIMARY CONTROLFILE FROM '${CONTROL_FILE}';"
  log_message "Restoring the primary control file ${CONTROL_FILE} for database sid ${ORACLE_SID}."
fi

# In case of Standalone and Active-Passive first start the database in no mount
# mode, restore the control file and then change database open_mode to mount.
# "sbt_channel_vec" and "sbt_clear_channel_vec will only be valid in case of SBT restore else will be empty.

if [ ! -z "${START_MOUNT_DB}" ]; then
  ${RMAN_PATH} >>${LOG_FILE} <<EOF
  connect target '${connect_string}${rman_privilege}';
  SET ECHO ON;
  SET DBID ${DATABASE_ID};
  STARTUP NOMOUNT;
  run {
    $sbt_channel_vec
    ${RESTORE_CLAUSE}
    $sbt_clear_channel_vec
  }
  STARTUP MOUNT;
EOF
# In case of RAC we will issue command like stop, startup mount, startup
# nomount using srvctl. In such case we will not do startup here.
# "sbt_channel_vec" and "sbt_clear_channel_vec will only be valid in case of SBT restore else will be empty.
else
  ${RMAN_PATH} >>${LOG_FILE} <<EOF
  connect target '${connect_string}${rman_privilege}';
  SET ECHO ON;
  SET DBID ${DATABASE_ID};
  run {
    $sbt_channel_vec
    ${RESTORE_CLAUSE}
    $sbt_clear_channel_vec
  }
EOF
fi

result=$?

if [ ${result} -eq 0 ];then
    log_message "***Operation completed on ${ORACLE_SID} at: $(date)"
else
    log_message "***Operation failed on ${ORACLE_SID} at: $(date)"
    log_message "Starting up the database in mount mode.";
    ${RMAN_PATH} >>${LOG_FILE} <<EOF
    connect target '${connect_string}${rman_privilege}';
    # Attempt to start the RMAN.
    startup mount
EOF
fi

exit ${result}
