#!/bin/bash
# Copyright 2021 Cohesity Inc.
#
# Author: Manipal Singh
#
# Referred from restore_to_original.rman file.
# This script performs the restore to a given time value.
# It takes command ID, Backuppiece Tag, Restore Point in Time
# and incarnation id as input.
# This script is invoked using rman_shell.sh script.

# Exit from the script if any of the step fails.
set -e

current_dir=$(dirname `readlink -f $0`)

SCRIPT_NAME=`basename "$0"`
# Source common_utils to use log_message.
. $current_dir/common_utils.sh

# Only parse options that are of interest to this recovery script.
# (getopts call below must contain every possible options. Reference $RMAN_SHELL_SCRIPT_OPTIONS
#  to get the full list of options)
#
while getopts "${RMAN_SHELL_SCRIPT_OPTIONS}" option; do
  case $option in
    C)
      PDB_LIST=$OPTARG;;
    D)
      db_incarnation=$OPTARG;;
    I)
      ID=$OPTARG;;
    P)
      POINT_IN_TIME=$OPTARG;;
    R)
      RMAN_PATH=$OPTARG;;
    T)
      TAG=$OPTARG;;
    n)
      NO_OPEN_MODE=1;;
    r)
      RMAN_BACKUP_TYPE=$OPTARG;;
    k)
      RESTORE_SCN=$OPTARG;;
    *) ;;
  esac
done

# Modify how to invoke Oracle binaries if the agent is installed with 'root'
setup_invoke_oracle_binaries

reset_incarnation_cmd=""
if [[ "x$db_incarnation" != "x" ]] ; then
  reset_incarnation_cmd="reset database to incarnation ${db_incarnation};"
fi
if [[ "x$PDB_LIST" != "x" ]] ; then
  # Do not set incarnation in case of PDB restore.
  reset_incarnation_cmd=""
fi

prerestore_crosscheck_cmd="";
if [ "${RMAN_BACKUP_TYPE}" != "${SBT_BACKUP}" ]; then
  prerestore_crosscheck_cmd="crosscheck backup;
  delete NOPROMPT expired backup;
  crosscheck datafilecopy all;
  delete NOPROMPT expired datafilecopy all;"
fi

close_pdb_cmd=""
if [[ "x$PDB_LIST" != "x" ]] ; then
  close_pdb_cmd="ALTER pluggable database ${PDB_LIST} close immediate force instances=all;"
fi

set_until_cmd=""
if [[ "x$POINT_IN_TIME" != "x" ]] ; then
  set_until_cmd="SET UNTIL TIME \"TO_DATE('${POINT_IN_TIME}','DD:MM:YYYY-HH24:MI:SS')\";"
fi

if [ ! -z $RESTORE_SCN ]; then
  log_message "Restoring database to SCN: $RESTORE_SCN instead of time: ${POINT_IN_TIME}";
  set_until_cmd="SET UNTIL SCN $RESTORE_SCN;"
fi

restore_cmd="RESTORE database;"
if [[ "x$PDB_LIST" != "x" ]]; then
  restore_cmd="RESTORE pluggable database '$PDB_LIST';"
fi

recovery_cmd="RECOVER database;"
if [[ "x$PDB_LIST" != "x" ]]; then
  recovery_cmd=""
  for pdb in $(echo $PDB_LIST | sed "s/,/ /g")
  do
    recovery_cmd="${recovery_cmd} recover pluggable database ${pdb}; alter pluggable database ${pdb} open resetlogs; alter pluggable database ${pdb} close; alter pluggable database ${pdb} open instances = all;"
  done
fi

open_cmd=""
if [[ "x$PDB_LIST" == "x" ]] && [ -z "${NO_OPEN_MODE}" ]; then
  open_cmd="ALTER DATABASE OPEN RESETLOGS;"
fi

postrestore_crosscheck_cmd="";
if [ "${RMAN_BACKUP_TYPE}" != "${SBT_BACKUP}" ]; then
  postrestore_crosscheck_cmd="crosscheck archivelog all;
  delete NOPROMPT expired archivelog all;
  crosscheck datafilecopy tag '$TAG';
  crosscheck backup tag '$TAG';
  delete NOPROMPT expired backup tag '$TAG';"
else
  postrestore_crosscheck_cmd="crosscheck archivelog all;"
fi

# We want to ignore error from prerestore block.
set +e
$RMAN_PATH msglog=${LOG_FILE} APPEND <<EOF
  connect target '${connect_string}${rman_privilege}';
  set echo on;
  set command id to '$ID';
  ${prerestore_crosscheck_cmd}
EOF

if [ $? -ne 0 ]; then
  log_message "Failure in prerestore crosscheck command, ignoring and proceeding with restore."
fi
set -e

$RMAN_PATH msglog=${LOG_FILE} APPEND <<EOF
  connect target '${connect_string}${rman_privilege}';
  set echo on;
  set command id to '$ID';
  ${reset_incarnation_cmd}
  list backup;
  list datafilecopy all;
  RUN {
    ${close_pdb_cmd}
    ${channel_vec}
    ${sbt_channel_vec}
    ${set_until_cmd}
    ${restore_cmd}
    ${recovery_cmd}
    ${open_cmd}
    ${sbt_channel_clear_vec}
    ${channel_clear_vec}
  }
EOF

err_code=$?
log_message "Recovery completed with exit status: ${err_code}"

# After this point, we will handle the errors ourselves.
set +e

# We can ignore the errors encountered in post restore crosscheck as there
# might not be enteries present corresponding to $TAG and it will cause
# unnecessary failure.
$RMAN_PATH msglog=${LOG_FILE} APPEND <<EOF
  connect target '${connect_string}${rman_privilege}';
  set echo on;
  set command id to '$ID';
  RUN {
    ${channel_vec}
    ${sbt_channel_vec}
    ${postrestore_crosscheck_cmd}
    ${sbt_channel_clear_vec}
    ${channel_clear_vec}
  }
EOF

exit $err_code
